# ## Setup

import os
import threading
import time
from typing import List
import modal

MODEL_NAME = "mixedbread-ai/mxbai-rerank-large-v1"
MODEL_DIR = f"/root/models/{MODEL_NAME}"
GPU_CONFIG = "L4:1"

APP_NAME = "ai2-scholar-qa"
APP_LABEL = APP_NAME.lower()


# ## Define a container image
#
# We want to create a Modal image which has the model weights pre-saved to a directory. The benefit of this
# is that the container no longer has to re-download the model from Huggingface - instead, it will take
# advantage of Modal's internal filesystem for faster cold starts.
#
# ### Download the weights
#
# We can download the model to a particular directory using the HuggingFace utility function `snapshot_download`.
#
# If you adapt this example to run another model,
# note that for this step to work on a [gated model](https://huggingface.co/docs/hub/en/models-gated)
# the `HF_TOKEN` environment variable must be set and provided as a [Modal Secret](https://modal.com/secrets).
#
# This can take some time -- at least a few minutes.

def download_model_to_image(model_dir, model_name):
    from huggingface_hub import snapshot_download
    from transformers.utils import move_cache

    os.makedirs(model_dir, exist_ok=True)

    snapshot_download(
        model_name,
        local_dir=model_dir,
    )
    move_cache()


# ### Image definition
#
# We’ll start from a basic Linux container image, install related libraries,
# and then use `run_function` to run the function defined above and ensure the weights of
# the model are saved within the container image.

reranker_image = (
    modal.Image.debian_slim(python_version="3.10")
    .pip_install(
        "torch==2.3.1",
        "transformers==4.44.1",
        "sentencepiece",
        "hf-transfer==0.1.6",
        "huggingface_hub==0.23.4",
        "sentence-transformers==3.0.1",
        "peft==0.13.2"
    )
    .env({"HF_HUB_ENABLE_HF_TRANSFER": "1"})
    .run_function(
        download_model_to_image,
        timeout=60 * 20,
        # secrets=[modal.Secret.from_name("chrisn-wildguard-hf-gated-read")],
        kwargs={
            "model_dir": MODEL_DIR,
            "model_name": MODEL_NAME,
        },
    )
    .add_local_python_source("custom_cross_encoder")
)

with reranker_image.imports():
    import torch
    from sentence_transformers import CrossEncoder
    from custom_cross_encoder import PaddedCrossEncoder
    from typing import List

app = modal.App(APP_NAME)


# ## The model class
#
# The inference function is best represented with Modal's [class syntax](/docs/guide/lifecycle-functions) and the `@enter` decorator.
# This enables us to load the model into memory just once every time a container starts up, and keep it cached
# on the GPU for each subsequent invocation of the function.
#
@app.cls(
    gpu=GPU_CONFIG,
    timeout=60 * 20,
    scaledown_window=60 * 20,
    min_containers=3,
    image=reranker_image
)
@modal.concurrent(max_inputs=1)
class Model:
    @modal.enter()
    def start_engine(self):
        print("🥶 cold starting inference")
        start = time.monotonic_ns()
        device = "cuda" if torch.cuda.is_available() else "cpu"
        self.reranker_compiled = PaddedCrossEncoder(
            MODEL_DIR,
            automodel_args={"torch_dtype": "float16"},
            trust_remote_code=True,
            device=device,
        )
        self.compiled_model = torch.compile(self.reranker_compiled.model, dynamic=True)
        self.reranker_compiled.model = self.compiled_model

        self.reranker_torch = CrossEncoder(
            MODEL_DIR,
            automodel_args={"torch_dtype": "float16"},
            trust_remote_code=True,
            device=device,
        )

        self.compiled_flag = False
        self.compiling_thread = None
        duration_s = (time.monotonic_ns() - start) / 1e9
        print(f"🏎️ engine started in {duration_s:.0f}s")

    def compile_reranker(self, batch_size):
        print("compiling cross encoder model")
        sentence_pairs = [["This is the query entered by the user.", f"This is candidate # {i}"] for i in range(batch_size)]
        for _ in range(2):
            self.reranker_compiled.predict(sentence_pairs, convert_to_tensor=True, show_progress_bar=True, batch_size=batch_size)
        self.compiled_flag = True
        print("compilation done!")

    @modal.method()
    def get_scores(self, query: str, passages: List[str], batch_size: int) -> List[float]:
        print(query)
        sentence_pairs = [[query, passage] for passage in passages]
        try:
            if self.compiled_flag:
                print("reranking with compiled model")
                scores = self.reranker_compiled.predict(sentence_pairs, convert_to_tensor=True, show_progress_bar=True,
                                               batch_size=batch_size).tolist()
            else:
                print("reranking with torch model")
                scores = self.reranker_torch.predict(sentence_pairs, convert_to_tensor=True, show_progress_bar=True,
                                                     batch_size=16).tolist()
                if self.compiling_thread is None:
                    self.compiling_thread = threading.Thread(target=self.compile_reranker, args=(batch_size,))
                    self.compiling_thread.start()
        except Exception as e:
            print(e)
            scores = self.reranker_torch.predict(sentence_pairs, convert_to_tensor=True, show_progress_bar=True,
                                           batch_size=batch_size).tolist()
        return [float(s) for s in scores]


# ## Coupling a frontend web application
#
# We can stream inference from a FastAPI backend, also deployed on Modal.


api_image = (
    modal.Image.debian_slim(python_version="3.11")
)


@app.function(
    image=api_image,
    min_containers=2,
    timeout=60 * 10,
)
@modal.concurrent(max_inputs=20)
async def inference_api(query: str, passages: List[str], batch_size: int = 512) -> List[float]:
    model = Model()
    return model.get_scores.remote(query, passages, batch_size)


@app.local_entrypoint()
def main():
    start = time.time()
    s = inference_api.remote("What is the python package infinity_emb?",
                             ["This is a document not related to the python package infinity_emb, hence...",
                              "Paris is in France!",
                              "infinity_emb is a package for sentence embeddings and rerankings using transformer models in Python!"],
                             64)
    print(s)
    print("Time taken:", time.time() - start)



SYSTEM_PROMPT = """
You are an intelligent and precise assistant that can understand the contents of research papers. 
You are knowledgable in different fields and domains of science, in particular computer science. 
You are able to interpret research papers, create questions and answers, and compare multiple papers.
"""

ATTRIBUTE_PROMPT = """
I want to write a related work section about following content:\n 
\"{}\".\n
We would like you to identify attributes that should be compared 
between related papers in this section. You will be provided with 
a list of the title of each paper and abstract. Your task is the following: 
Given a list of papers, you should identify and define {} attributes 
that can be used to compare the given papers within the given subsection topic. 
Attributes should focus on the main content of the papers, and should **NOT** be about their metadata such as Authors, Publication Year, or Venues.
Return a JSON object in the following format:\n{{\"attribute name\": {{\"def\": \"definition for comparable attribute within the given subsection topic\", \"is_metadata\": \"True or False\"}}, \"attribute name\": {{\"def\": \"definition for comparable attribute within the given subsection topic\", \"is_metadata\": \"True or False\"}}, ...}}
\n\n{}\n",
"""

VALUE_GENERATION_FROM_ABSTRACT = """
Imagine the following scenario: A user is making a table for a scholarly paper that contains information about multiple papers and compares these papers. 
To compare and contrast the papers, the user has selected some aspects by which papers should be compared. 
Your task is the following: Given a paper abstract and a question describing one of the selected aspects, generate a value that can be added to the table. 
You should find the part of the abstract that discusses the aspect provided in the question.
If there is no answer in the abstract, return \"N/A\" as the value to be added. 
**Ensure that you follow these rules: (1) Only return the answer. Do not repeat the question or add any surrounding text. (2) The answer should be brief and consist of phrases of fewer than 10 words.**\n\n
"""

VALUE_GENERATION_FROM_METADATA = """
Imagine the following scenario: A user is making a table for a scholarly paper that contains information about multiple papers and compares these papers.
The user wants to some metadata about each paper to the table.
Your task is the following: Given a json blob containing all available metadata (e.g., title, abstract, authors, etc.) for a paper, 
generate a value that can be added to the table under the following column: {}.
**Ensure that you follow these rules: (1) Only return the answer. Do not repeat the question or add any surrounding text. (2) The answer should be brief and consist of phrases of fewer than 10 words. (3) Return N/A if there is no answer.**\n\n
"""

VALUE_CONSISTENCY_PROMPT_ZS = """
Imagine the following scenario: A user is making a table for a scholarly paper that contains information about multiple papers and compares these papers. 
To compare and contrast the papers, the user has selected an aspect which will be added as a column to the table. 
Your task is the following: Given the column name and information from each paper relevant to that column, generate final values to be added to the table.
Return the output as a JSON object in the following format:\n{{\"values\": [\"value for paper 1\", \"value for paper 2\", ...]}}\n
**Ensure that you follow these rules: (1) None of the values provided in the prompt should be overwritten by "N/A" (2) Only return a single JSON object. (3) JSON object should be complete and valid. (4) The list in the JSON object should contain the same number of values as provided in the input (5) Each input should have a corresponding output value (leave "N/A" values unchanged) (6) All values should follow consistent formatting and style.\n\n**
"""

VALUE_CONSISTENCY_PROMPT_FS = """
Imagine the following scenario: A user is making a table for a scholarly paper that contains information about multiple papers and compares these papers. 
To compare and contrast the papers, the user has selected an aspect which will be added as a column to the table. 
Your task is the following: Given the column name and information from each paper relevant to that column, generate final values to be added to the table.
Return the output as a JSON object in the following format:\n{{\"values\": [\"value for paper 1\", \"value fpr paper 2\", ...]}}\n
**Ensure that you follow these rules: (1) None of the values provided in the prompt should be overwritten by "N/A" (2) Only return a single JSON object. (3) JSON object should be complete and valid. (4) The list in the JSON object should contain the same number of values as provided in the input (5) Each input should have a corresponding output value (leave "N/A" values unchanged) (6) All values should follow consistent formatting and style.\n\n**
"""

VESPAQA_PROMPT = """
Answer a question using the provided relevant snippets from a scientific paper.
Your response should be a JSON object with the following fields:
  - answer: The answer to the question. The answer should use concise language, but be comprehensive. Only provide answers that are objectively supported by the text in paper. Avoid answers longer than 20 words. If you cannot answer the question using provided information, the answer should be 'N/A.
  - excerpts: A list of one or more *EXACT* text spans extracted from the paper that support the answer. Return between at most five spans, and no more that 500 words. Make sure to cover all aspects of the answer above. If the answer is 'N/A', excerpts should be an empty list.

<Start of snippets>

Title: [TITLE]

[SNIPPETS]

<End of snippets>

Given the information above, please answer the question: [QUESTION] \nAnswer:"""


from datetime import datetime
from typing import List, Dict, Any

from scholarqa.config.config_setup import LogsConfig
from scholarqa.llms.constants import CostAwareLLMResult
from scholarqa.models import ToolRequest
from scholarqa.trace.trace_writer import GCSWriter, LocalWriter


class EventTrace:
    def __init__(self, task_id: str, n_retrieval: int, n_rerank: int, req: ToolRequest, user_id: str = None):
        self.query = req.query
        if not user_id:
            try:
                self.user_id = req.user_id
            except Exception as e:
                self.user_id = None
        else:
            self.user_id = user_id
        self.task_id = task_id
        self.timestamp = datetime.now().isoformat()
        self.n_retrieval = n_retrieval
        self.n_retrieved = 0
        self.n_candidates = 0
        self.n_rerank = n_rerank
        self.opt_in = req.opt_in
        self.total_cost = 0.0

        self.decomposed_query = dict()
        self.candidates = []
        self.retrieved = []
        self.quotes = {"cost": 0.0, "quotes": []}
        self.cluster = dict()
        self.summary = dict()
        self.total_cost = 0.0
        self.tokens = {"input": 0, "output": 0, "total": 0, "reasoning": 0}

    def trace_decomposition_event(self, decomposed_query: CostAwareLLMResult):
        self.decomposed_query = decomposed_query.result._asdict()
        self.decomposed_query["cost"] = decomposed_query.tot_cost
        self.decomposed_query["model"] = decomposed_query.models[0]
        self.decomposed_query["tokens"] = decomposed_query.tokens._asdict()
        self.total_cost += decomposed_query.tot_cost
        for k,v in self.tokens.items():
            self.tokens[k] += self.decomposed_query["tokens"][k]

    def trace_retrieval_event(self, retrieved: List[Dict[str, Any]]):
        self.n_retrieved = len(retrieved)
        self.retrieved = retrieved

    def trace_rerank_event(self, candidates: List[Dict[str, Any]]):
        self.n_candidates = len(candidates)
        self.candidates = candidates

    def trace_quote_event(self, paper_summaries: CostAwareLLMResult):
        topk = [{"idx": i, "key": k, "snippets": v} for
                i, (k, v) in enumerate(paper_summaries.result.items())]
        topk_models = paper_summaries.models
        for idx, tk in enumerate(topk):
            tk["model"] = topk_models[idx]

        self.quotes["cost"] = paper_summaries.tot_cost
        self.quotes["tokens"] = paper_summaries.tokens._asdict()
        self.quotes["quotes"] = topk
        self.total_cost += paper_summaries.tot_cost
        for k, v in self.tokens.items():
            self.tokens[k] += self.quotes["tokens"][k]

    def trace_clustering_event(self, cluster_json: CostAwareLLMResult, plan_str: Dict[str, Any]):
        self.cluster["cost"] = cluster_json.tot_cost
        self.cluster["tokens"] = cluster_json.tokens._asdict()
        self.cluster["cot"] = cluster_json.result["cot"]
        self.cluster["plan"] = plan_str
        self.cluster["model"] = cluster_json.models[0]
        self.total_cost += cluster_json.tot_cost
        for k, v in self.tokens.items():
            self.tokens[k] += self.cluster["tokens"][k]

    def trace_inline_citation_following_event(self, paper_summaries_extd: Dict[str, Any], quotes_metadata: Dict[str, List[Dict[str, Any]]]):
        for quote_obj in self.quotes["quotes"]:
            quote_obj["snippets"] = paper_summaries_extd[quote_obj["key"]].get("quote", quote_obj["snippets"])
            quote_obj["inline_citations"] = paper_summaries_extd[quote_obj["key"]].get("inline_citations", dict())
            quote_obj["metadata"] = quotes_metadata.get(quote_obj["key"], [])

    def trace_summary_event(self, json_summary: List[Dict[str, Any]], cost_result: CostAwareLLMResult, tab_costs: List[Dict] = None):
        self.summary = {"sections": json_summary, "cost": cost_result.tot_cost, "tokens": cost_result.tokens._asdict(), "table_costs": tab_costs}
        for idx, section in enumerate(self.summary["sections"]):
            section["model"] = cost_result.models[idx]
        self.total_cost += cost_result.tot_cost
        for k, v in self.tokens.items():
            self.tokens[k] += self.summary["tokens"][k]

        if tab_costs:
            for tcost in tab_costs:
                column_cost = tcost.get("column_cost", 0.0)
                if column_cost:
                    self.total_cost += column_cost["cost_value"]
                    self.tokens["input"] += column_cost["tokens"]["prompt"]
                    self.tokens["output"] += column_cost["tokens"]["completion"]
                    self.tokens["total"] += column_cost["tokens"]["total"]
                    self.tokens["reasoning"] += column_cost["tokens"].get("reasoning", 0)

                cell_cost = tcost.get("cell_cost", 0.0)
                if cell_cost:
                    for ccost in cell_cost:
                        if not isinstance(ccost, dict):
                            continue
                        for v in ccost.values():
                            self.total_cost += v["cost_value"]
                            self.tokens["input"] += v["tokens"]["prompt"]
                            self.tokens["output"] += v["tokens"]["completion"]
                            self.tokens["total"] += v["tokens"]["total"]
                            self.tokens["reasoning"] += v["tokens"].get("reasoning", 0)


    def persist_trace(self, logs_config: LogsConfig):
        trace_writer = GCSWriter(bucket_name=logs_config.event_trace_loc) if logs_config.tracing_mode == "gcs" \
            else LocalWriter(local_dir=f"{logs_config.log_dir}/{logs_config.event_trace_loc}")
        trace_writer.write(trace_json=self, file_name=self.task_id)



from abc import ABC, abstractmethod
from typing import List, Dict, Any

import logging

logger = logging.getLogger(__name__)

try:
    import torch
    import torch.nn.functional as F
except ImportError:
    logger.warning("torch not found, custom baseline rerankers will not work.")


class AbstractReranker(ABC):
    @abstractmethod
    def get_scores(self, query: str, documents: List[str]) -> List[float]:
        pass


class SentenceTransformerEncoder:
    def __init__(self, model_name_or_path: str):
        from sentence_transformers import SentenceTransformer
        device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model = SentenceTransformer(model_name_or_path, revision=None, device=device)

    def encode(self, sentences: List[str]):
        return self.model.encode(sentences, show_progress_bar=True, convert_to_tensor=True)

    def get_tokenizer(self):
        return self.model.tokenizer


# GIST embeddings model supported by Sentence transformer
# https://huggingface.co/avsolatorio/GIST-large-Embedding-v0
class BiEncoderScores(AbstractReranker):
    def __init__(self, model_name_or_path: str):
        self.model = SentenceTransformerEncoder(model_name_or_path)

    def get_scores(self, query: str, passages: List[str]) -> List[float]:
        query_embedding = self.model.encode([query])[0]
        passage_embeddings = self.model.encode(passages)
        scores = F.cosine_similarity(query_embedding.unsqueeze(0), passage_embeddings).cpu().numpy()
        return [float(s) for s in scores]


# Sentence Transformer supports Jina AI (https://huggingface.co/jinaai/jina-reranker-v2-base-multilingual)
# and Mix Bread re-rankers (https://huggingface.co/mixedbread-ai/mxbai-rerank-large-v1)
class CrossEncoderScores(AbstractReranker):
    def __init__(self, model_name_or_path: str):
        from sentence_transformers import CrossEncoder
        device = "cuda" if torch.cuda.is_available() else "cpu"
        print(device)
        self.model = CrossEncoder(
            model_name_or_path,
            automodel_args={"torch_dtype": "float16"},
            trust_remote_code=True,
            device=device,
        )

    def get_tokenizer(self):
        return self.model.tokenizer

    def get_scores(self, query: str, passages: List[str]) -> List[float]:
        sentence_pairs = [[query, passage] for passage in passages]
        scores = self.model.predict(sentence_pairs, convert_to_tensor=True, show_progress_bar=True,
                                    batch_size=128).tolist()
        return [float(s) for s in scores]


# Supports the BAAI/bge... models https://huggingface.co/BAAI/bge-reranker-v2-m3
class FlagEmbeddingScores:
    def __init__(self, model_name_or_path: str):
        from FlagEmbedding import FlagReranker
        self.model = FlagReranker(model_name_or_path, use_fp16=True)

    def get_tokenizer(self):
        return self.model.tokenizer

    def get_scores(self, query: str, passages: List[str], separator: str) -> List[float]:
        sentence_pairs = [(query, passage.replace(separator, self.get_tokenizer().eos_token)) for passage in passages]
        scores = self.model.compute_score(sentence_pairs, normalize=True, batch_size=32)
        return [float(s) for s in scores]

RERANKER_MAPPING = {
    "crossencoder": CrossEncoderScores,
    "biencoder": BiEncoderScores,
    "flag_embedding": FlagEmbeddingScores
}


from typing import List, Dict, Optional
import uuid

import itertools
from concurrent.futures import ThreadPoolExecutor

from scholarqa.table_generation.table_model import TableWidget, TableColumn, TableRow, TableCell
from scholarqa.table_generation.column_suggestion import generate_attribute_suggestions
from scholarqa.table_generation.value_generation import generate_value_suggestions
from scholarqa.utils import get_paper_metadata
from scholarqa.rag.retrieval import PaperFinder
from scholarqa.llms.constants import GPT_4o
from scholarqa.llms.litellm_helper import CostAwareLLMCaller, CostReportingArgs

class TableGenerator:
    def __init__(
        self,
        paper_finder: PaperFinder,
        llm_caller: CostAwareLLMCaller,
        max_threads: int = 1,
    ) -> None:
        self.paper_finder = paper_finder
        self.llm_caller = llm_caller
        self.max_threads = max_threads
        self.empty_cell = TableCell(id="empty", value="N/A", display_value="N/A", metadata={})
    
    def run_table_generation(
        self, 
        thread_id: str,
        user_id: str,
        original_query: str, 
        section_title: str, 
        corpus_ids: List[int],
        column_num: int = 10,
        run_subselection: bool = True,
        column_model: Optional[str] = GPT_4o,
        value_model: Optional[str] = GPT_4o,
    ) -> TableWidget:
        """
        Entry point to generate a complete table, given the original
        query sent by the user to ScholarQA, the title of the section
        under which the table will be displayed and corpus IDs of all
        papers to be included (i.e., ones cited in the section).
        """

        # Step 1: Construct a query for the column suggestion tool using
        # the section title and original user query as input. Also create
        # a cost argument object to allow the the tool to track costs
        column_suggestion_query = f"{section_title}"
        if original_query != "":
            column_suggestion_query += f", a section included in an answer to the question: {original_query}"
        cost_args = CostReportingArgs(
            task_id=thread_id,
            user_id=user_id,
            msg_id=thread_id,
            description="Cost for generating column suggestions",
            model=column_model,
        )
        output = generate_attribute_suggestions(
            corpus_ids=[str(x) for x in corpus_ids],
            query=column_suggestion_query,
            model=column_model,
            llm_caller=self.llm_caller,
            column_num=column_num,
        )
        column_cost = output.get("cost", {})
        
        # Step 2: Create a new table data structure with suggested columns.
        # While creating a column, also create requests to call the value
        # generation functionality for each column.  
        table = TableWidget(id=thread_id)
        column_suggestions = output.get("columns", [])
        value_gen_requests = []
        for column in column_suggestions:
            # Generate a UUID for the column
            column_id = str(uuid.uuid4())
            # Sometimes column names have underscores - replace them for readability
            column_name = column['name'].replace('_', ' ').title()
            table.add_columns(
                [TableColumn(
                    id=column_id,
                    name=column_name,
                    description=column["definition"],
                    is_metadata=column["is_metadata"],
                    tools=["table_cell_value_generation"],
                )]
            )
            value_gen_requests.append({
                "column_id": column_id,
                "column_name": column_name,
                "column_def": column["definition"],
                "corpus_ids": [str(x) for x in corpus_ids],
                "is_metadata": column["is_metadata"],
                "model": value_model,
                "paper_finder": self.paper_finder,
                "llm_caller": self.llm_caller,
                "cost_args": CostReportingArgs(
                    task_id=thread_id,
                    user_id=user_id,
                    msg_id=thread_id,
                    description=f"Cost for generating cell values for column {column_name}",
                    model=value_model,
                ),
            })
        
        # Step 3: Since we have corpus IDs, get paper titles and them as rows in the first
        # column of the table. Store the correspondence between corpus IDs and row IDs 
        # to use for cell creation later on.
        paper_info = self.retrieve_paper_info(corpus_ids)
        row_id_map = {}
        for corpus_id in corpus_ids:
            row_id = str(uuid.uuid4())
            row_id_map[corpus_id] = row_id
            table.add_rows([
                TableRow(
                    id=row_id,
                    display_value=paper_info[corpus_id]["title"],
                    paper_corpus_id=corpus_id,
                )
            ])

        # Step 4: Run value generation requests for all columns in parallel and add cells to the table
        all_cell_costs = []
        with ThreadPoolExecutor(max_workers=self.max_threads) as executor:
            output = list(executor.map(
                self.generate_values,  
                itertools.repeat(row_id_map),
                value_gen_requests,
            ))
            for item in output:
                new_cells = item.get("cells", {})
                cell_costs = item.get("cost", {})
                table.cells.update(new_cells)
                all_cell_costs.append(cell_costs)

        if run_subselection:
            table = self.subselect_columns_and_rows(table)
        final_cost_dict = {
            "column_cost": column_cost,
            "cell_cost": all_cell_costs,
        }

        return table, final_cost_dict
    
    """
    Functions to only select a subset of informative 
    and non-redundant columns and rows from the table
    """
    def column_to_doc(self, column_id: str, table: TableWidget):
        cell_ids = [f"{row.id}_{column_id}" for row in table.rows]
        cells = [table.cells.get(cell_id, self.empty_cell).value for cell_id in cell_ids]
        cells = [cell for cell in cells if (cell != None and len(cell) > 0 and cell != "N/A")]
        return {'valid_cells': len(cells), 'doc': ' ||| '.join(cells)}
        # return {'valid_cells': len(cells), 'doc': ' ||| '.join(map(clean, cells))}

    def row_to_doc(self, row_id: str, table: TableWidget):
        cell_ids = [f"{row_id}_{column.id}" for column in table.columns]
        cells = [table.cells.get(cell_id, self.empty_cell).value for cell_id in cell_ids]
        cells = [cell for cell in cells if (cell != None and len(cell) > 0 and cell != "N/A")]
        return {'valid_cells': len(cells), 'doc': ' ||| '.join(cells)}
        # return {'valid_cells': len(cells), 'doc': ' ||| '.join(map(clean, cells))}

    def keep_rows(self, table: TableWidget, row_ids: List[str]):
        table.rows = [row for row in table.rows if row.id in row_ids]
        table.cells = {cell_id: cell for cell_id, cell in table.cells.items() if cell_id.split('_')[0] in row_ids}
        return table

    def keep_columns(self, table: TableWidget, column_ids: List[str]):
        table.columns = [column for column in table.columns if column.id in column_ids]
        table.cells = {cell_id: cell for cell_id, cell in table.cells.items() if cell_id.split('_')[1] in column_ids}
        return table


    def subselect_columns_and_rows(self, original_table: TableWidget, max_rows=6, max_columns=6):
        table = original_table.model_copy(deep=True)

        row_valid_cells = [{'row_id': row.id, 'valid_cells': self.row_to_doc(row.id, table)['valid_cells']} for row in table.rows]
        row_valid_cells = [row for row in row_valid_cells if row['valid_cells'] >= max_columns]
        row_valid_cells = sorted(row_valid_cells, key=lambda x: x['valid_cells'], reverse=True)
        row_valid_cells = row_valid_cells[:max_columns*2]
        
        table = self.keep_rows(table, [row['row_id'] for row in row_valid_cells])

        column_valid_cells = [{'column_id': column.id, **self.column_to_doc(column.id, table)} for column in table.columns]
        column_valid_cells = [column for column in column_valid_cells if column['valid_cells'] > len(table.rows)*0.7]

        column_valid_cells = column_valid_cells[:max_columns]

        table =self.keep_columns(table, [column['column_id'] for column in column_valid_cells])

        row_valid_cells = [{'row_id': row.id, 'valid_cells': self.row_to_doc(row.id, table)['valid_cells']} for row in table.rows]
        row_valid_cells = [row for row in row_valid_cells if row['valid_cells'] >= max_rows / 2]
        row_valid_cells = sorted(row_valid_cells, key=lambda x: x['valid_cells'], reverse=True)
        row_valid_cells = row_valid_cells[:max_rows]
        
        table = self.keep_rows(table, [row['row_id'] for row in row_valid_cells])
        return table

    def retrieve_paper_info(self, corpus_ids: List[str]) -> Dict:
        """ 
        Given a set of corpus IDs for papers to be added to the table,
        retrieve paper titles using the Semantic Scholar batch querying API.
        """
        paper_metadata = get_paper_metadata(corpus_ids)
        paper_info = {corpus_id: paper_metadata[str(corpus_id)] if str(corpus_id) in paper_metadata else {} for corpus_id in corpus_ids}
        return paper_info

    def generate_values(self, row_id_map: dict, request: dict):
        """
        Given a request to generate cell values for a column, call the value 
        generation functionality and create and return TableCell objects for each.
        """
        column_id = request.pop("column_id")
        output = generate_value_suggestions(**request)
        generated_values = output.get("cell_values", [])
        cell_costs = output.get("cost", {})
        table_cells = {}
        for value in generated_values:
            cell_id = f"{row_id_map[int(value['corpusId'])]}_{column_id}"
            cell = TableCell(
                id=cell_id,
                value=value['displayValue'],
                display_value=value['displayValue'],
                metadata=value.get('metadata', None),
            )
            table_cells[cell_id] = cell
        output ={
            "cells": table_cells,
            "cost": cell_costs, 
        }
        return output


import logging
import multiprocessing
import os
from json import JSONDecodeError
from time import time
from typing import Union
from uuid import uuid4, uuid5, UUID

from fastapi import FastAPI, HTTPException, Request
from nora_lib.tasks.models import TASK_STATUSES, AsyncTaskState
from nora_lib.tasks.state import NoSuchTaskException

from scholarqa.config.config_setup import read_json_config
from scholarqa.models import (
    AsyncToolResponse,
    TaskResult,
    ToolRequest,
    ToolResponse,
    TaskStep
)
from scholarqa.rag.reranker.modal_engine import ModalReranker
from scholarqa.rag.reranker.reranker_base import RERANKER_MAPPING
from scholarqa.rag.retrieval import PaperFinderWithReranker, PaperFinder
from scholarqa.rag.retriever_base import FullTextRetriever
from scholarqa.scholar_qa import ScholarQA
from scholarqa.state_mgmt.local_state_mgr import LocalStateMgrClient
from typing import Type, TypeVar

logger = logging.getLogger(__name__)

TIMEOUT = 240

async_context = multiprocessing.get_context("fork")

started_task_step = None

T = TypeVar("T", bound=ScholarQA)


def lazy_load_state_mgr_client():
    return LocalStateMgrClient(logs_config.log_dir, "async_state")


def lazy_load_scholarqa(task_id: str, sqa_class: Type[T] = ScholarQA, **sqa_args) -> T:
    retriever = FullTextRetriever(**run_config.retriever_args)
    if run_config.reranker_args:
        reranker = RERANKER_MAPPING[run_config.reranker_service](**run_config.reranker_args)
        paper_finder = PaperFinderWithReranker(retriever, reranker, **run_config.paper_finder_args)
    else:
        paper_finder = PaperFinder(retriever, **run_config.paper_finder_args)

    return sqa_class(paper_finder=paper_finder, task_id=task_id, state_mgr=app_config.state_mgr_client,
                     logs_config=logs_config, **run_config.pipeline_args, **sqa_args)


# setup logging config and local litellm cache
CONFIG_PATH = os.environ.get("CONFIG_PATH", "run_configs/default.json")

app_config = read_json_config(CONFIG_PATH)
logs_config = app_config.logs
run_config = app_config.run_config
app_config.load_scholarqa = lazy_load_scholarqa


def _do_task(tool_request: ToolRequest, task_id: str) -> TaskResult:
    """
    TODO: BYO logic here. Don't forget to define `ToolRequest` and `TaskResult`
    in `models.py`!

    The meat of whatever it is your tool or task agent actually
    does should be kicked off in here. This will be run synchonrously
    unless `_needs_to_be_async()` above returns True, in which case
    it will be run in a background process.

    If you need to update state for an asynchronously running task, you can
    use `task_state_manager.read_state(task_id)` to retrieve, and `.write_state()`
    to write back.
    """
    scholar_qa = app_config.load_scholarqa(task_id)
    return scholar_qa.run_qa_pipeline(tool_request)


def _estimate_task_length(tool_request: ToolRequest) -> str:
    """

    For telling the user how long to wait before asking for a status
    update on async tasks. This can just be a static guess, but you
    have access to the request if you want to do something fancier.
    """
    return (
        "~3 minutes"
    )


###########################################################################
### BELOW THIS LINE IS ALL TEMPLATE CODE THAT SHOULD NOT NEED TO CHANGE ###
###########################################################################


def create_app() -> FastAPI:
    app = FastAPI(root_path="/api")

    @app.get("/")
    def root(request: Request):
        return {"message": "Hello World", "root_path": request.scope.get("root_path")}

    @app.get("/health", status_code=204)
    def health():
        return "OK"

    @app.post("/query_corpusqa")
    def use_tool(
            tool_request: ToolRequest,
    ) -> Union[AsyncToolResponse, ToolResponse]:
        if not app_config.state_mgr_client:
            app_config.state_mgr_client = lazy_load_state_mgr_client()
        # Caller is asking for a status update of long-running request
        if tool_request.task_id:
            return _handle_async_task_check_in(tool_request)

        # New task
        task_id = str(uuid4())
        logs_config.task_id = task_id
        logger.info("New task")
        app_config.state_mgr_client.init_task(task_id, tool_request)
        estimated_time = _start_async_task(task_id, tool_request)

        return AsyncToolResponse(
            task_id=task_id,
            query=tool_request.query,
            estimated_time=estimated_time,
            task_status=TASK_STATUSES["STARTED"],
            task_result=None,
            steps=[started_task_step]
        )
    app.state.use_tool_fn = use_tool
    return app


def _start_async_task(task_id: str, tool_request: ToolRequest) -> str:
    global started_task_step
    estimated_time = _estimate_task_length(tool_request)
    tool_request.task_id = task_id
    task_state_manager = app_config.state_mgr_client.get_state_mgr(tool_request)
    started_task_step = TaskStep(description=TASK_STATUSES["STARTED"], start_timestamp=time(),
                                 estimated_timestamp=time() + TIMEOUT)
    task_state = AsyncTaskState(
        task_id=task_id,
        estimated_time=estimated_time,
        task_status=TASK_STATUSES["STARTED"],
        task_result=None,
        extra_state={"query": tool_request.query, "start": time(),
                     "steps": [started_task_step]},
    )
    task_state_manager.write_state(task_state)

    def _do_task_and_write_result():
        extra_state = {}
        try:
            task_result = _do_task(tool_request, task_id)
            task_status = TASK_STATUSES["COMPLETED"]
            extra_state["end"] = time()
        except Exception as e:
            task_result = None
            task_status = TASK_STATUSES["FAILED"]
            extra_state["error"] = str(e)

        state = task_state_manager.read_state(task_id)
        state.task_result = task_result
        state.task_status = task_status
        state.extra_state.update(extra_state)
        state.estimated_time = "--"
        task_state_manager.write_state(state)

    async_context.Process(
        target=_do_task_and_write_result,
        name=f"Async Task {task_id}",
        args=(),
    ).start()

    return estimated_time


def _handle_async_task_check_in(
        tool_req: ToolRequest,
) -> Union[ToolResponse | AsyncToolResponse]:
    """
    For tasks that will take a while to complete, we issue a task id
    that can be used to request status updates and eventually, results.

    This helper function is responsible for checking the state store
    and returning either the current state of the given task id, or its
    final result.
    """
    task_id = tool_req.task_id
    logs_config.task_id = task_id
    task_state_manager = app_config.state_mgr_client.get_state_mgr(tool_req)
    try:
        task_state = task_state_manager.read_state(task_id)
    except NoSuchTaskException:
        raise HTTPException(
            status_code=404, detail=f"Referenced task {task_id} does not exist."
        )
    except JSONDecodeError as e:
        logger.warning("state file is corrupted, should be updated on next poll: {e}")
        return AsyncToolResponse(
            task_id=task_id,
            query="",
            estimated_time="~3 minutes",
            task_status=f"{time()}: Processing user query",
            task_result=None,
        )

    # Retrieve data, which is just on local disk for now
    if task_state.task_status == TASK_STATUSES["FAILED"]:
        if task_state.extra_state and "error" in task_state.extra_state:
            msg = f"\nError: {task_state.extra_state['error']}"
            logger.exception(msg)
        else:
            msg = f"Referenced task failed."
        raise HTTPException(status_code=500, detail=f"{msg}")

    if task_state.task_status == TASK_STATUSES["COMPLETED"]:
        if not task_state.task_result:
            msg = f"Task marked completed but has no result."
            logger.error(msg)
            raise HTTPException(
                status_code=500,
                detail=msg,
            )

        if "start" in task_state.extra_state and "end" in task_state.extra_state:
            try:
                cost = task_state.task_result["cost"] if type(
                    task_state.task_result) == dict else task_state.task_result.cost
            except Exception as e:
                logger.warning(f"Error occurred while parsing cost from the response: {e}")
                cost = 0.0
            logger.info(
                f"completed in {task_state.extra_state['end'] - task_state.extra_state['start']} seconds, "
                f"cost: ${cost}")
        return ToolResponse(
            task_id=task_state.task_id,
            query=task_state.extra_state["query"],
            task_result=task_state.task_result,
        )

    if task_state.task_status not in {TASK_STATUSES["COMPLETED"],
                                      TASK_STATUSES["FAILED"]} and "start" in task_state.extra_state:
        elapsed = time() - task_state.extra_state["start"]
        if elapsed > TIMEOUT:
            task_state.task_status = TASK_STATUSES["FAILED"]
            task_state.extra_state["error"] = f"Task timed out after {TIMEOUT} seconds"
            task_state_manager.write_state(task_state)
            logger.info(f"timed out after {time() - task_state.extra_state['start']} seconds.")
            raise HTTPException(
                status_code=500,
                detail=f"Task timed out after {TIMEOUT} seconds.")

    return AsyncToolResponse(
        task_id=task_state.task_id,
        query=task_state.extra_state["query"],
        estimated_time=task_state.estimated_time,
        task_status=task_state.task_status,
        task_result=task_state.task_result,
        steps=task_state.extra_state.get("steps", []),
    )



import os
import json
import requests
import time
import logging

from typing import List, Dict
from pydantic import BaseModel
import itertools
from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy

from scholarqa.table_generation.prompts import *
from scholarqa.utils import get_paper_metadata
from scholarqa.llms.constants import *
from scholarqa.rag.retrieval import PaperFinder
from scholarqa.llms.litellm_helper import CostAwareLLMCaller, CostReportingArgs, llm_completion

logger = logging.getLogger(__name__)

class PaperQAAnswer(BaseModel):
    answer: str
    exceprts: List[str]

def get_cost_object(completion: CompletionResult) -> dict:
    cost_dict = {
        "cost_value": completion.cost,
        "tokens": {
            "total": completion.total_tokens,
            "prompt": completion.input_tokens,
            "completion": completion.output_tokens,
            "reasoning": completion.reasoning_tokens
        },
        "model": completion.model,
    }
    return cost_dict

def get_metadata_columns(
        question: str, 
        metadata: dict, 
        model: str,
        llm_caller: CostAwareLLMCaller = None,
        cost_args: CostReportingArgs = None,
    ):
    """
    Given a question and metadata from a research paper, prompt
    an LLM to answer the question using the metadata provided. 
    We use this to populate metadata columns in tables (e.g., venue).
    """
    prompt = VALUE_GENERATION_FROM_METADATA.format(question)
    prompt += f"Metadata: {metadata}"
    cur_cost_args = deepcopy(cost_args)
    corpus_id = metadata.get("corpusId", None)
    cur_cost_args = CostReportingArgs(
        task_id=cost_args.task_id,
        user_id=cost_args.user_id,
        msg_id=cost_args.msg_id,
        description=cost_args.description + f" for corpus ID {corpus_id}",
        model=cost_args.model,
    )
    value_generation_params = {
        "user_prompt": prompt,
        "system_prompt": SYSTEM_PROMPT,
        "model": model,
        "fallback": GPT_4o,
    }
    output = llm_caller.call_method(
        cost_args=cur_cost_args,
        method=llm_completion,
        **value_generation_params,
    )
    response = output.result.content
    cost_dict = get_cost_object(output.result)
    response_simplified = {
        "question": question,
        "answer": response,
        "corpusId": metadata.get("corpusId", None),
        "source": "metadata",
        "evidenceId": None,
        "cost": cost_dict,
    }
    return response_simplified


def get_value_from_abstract(
        question: str, 
        corpus_id: str, 
        model: str,
        llm_caller: CostAwareLLMCaller = None,
        cost_args: CostReportingArgs = None,
    ):
    """
    Given a query and a paper's corpus ID, retrieve an answer
    to the query based on the paper abstract. We use this as a
    backoff strategy for papers without full-text access.
    """
    # Step 1: Retrieve abstract for provided corpus ID from Semantic Scholar API
    response = None
    retry_num = 0
    while response is None:
        try:
            response = get_paper_metadata([corpus_id])
        except Exception as e:
            logger.error(f"Error while retrieving paper metadata for corpus ID {corpus_id}: {str(e)}")
        retry_num += 1
        time.sleep(retry_num * 5)
    response_content = response[corpus_id]
    title = response_content["title"] if "title" in response_content else None
    abstract = response_content["abstract"] if "abstract" in response_content and response_content["abstract"] else None
    # Step 2: Prompt LLM to produce a cell value using the paper abstract
    prompt = VALUE_GENERATION_FROM_ABSTRACT + f"Paper title:{title}\nPaper abstract: {abstract}\nQuestion: {question}\nAnswer:"
    cur_cost_args = CostReportingArgs(
        task_id=cost_args.task_id,
        user_id=cost_args.user_id,
        msg_id=cost_args.msg_id,
        description=cost_args.description + f" for corpus ID {corpus_id}",
        model=cost_args.model,
    )
    value_generation_params = {
        "user_prompt": prompt,
        "system_prompt": SYSTEM_PROMPT,
        "model": model,
        "fallback": GPT_4o,
    }
    output = llm_caller.call_method(
        cost_args=cur_cost_args,
        method=llm_completion,
        **value_generation_params,
    )
    value = output.result.content
    cost_dict = get_cost_object(output.result)
    return value, cost_dict


def run_paper_qa(
        question: str, 
        corpus_id: str, 
        model: str,
        paper_finder: PaperFinder = None,
        llm_caller: CostAwareLLMCaller = None,
        cost_args: CostReportingArgs = None,
    ):
    """
    Given a query and a paper's corpus ID, retrieve an answer
    to the query from the paper full-text. This function relies
    on Vespa snippet search utility to first retrieve relevant
    passages from the paper full-text to produce answers and 
    evidence snippets backing them. If we are unable to find
    relevant snippets, we back off to generating an answer from
    the paper's abstract.
    """
    try:
        # Restrict snippet search only to the paper we're currently 
        # generating values for. Also drop formatting instructions
        # from the question for the retrieval function.
        filter_kwargs = {
            "paperIds" : f"CorpusId:{corpus_id}"
        }
        snippets = paper_finder.retrieve_passages(
            query=question.split("Only return the answer. ")[0], 
            **filter_kwargs,
        )
        if snippets:
            paper_title = snippets[0]["title"]
            concatenated_snippets = ""
            for i, snippet in enumerate(snippets):
                concatenated_snippets += f"Snippet {i+1}: {snippet['text']}\n\n"
            prompt = VESPAQA_PROMPT.replace('[TITLE]', paper_title)
            prompt = prompt.replace('[SNIPPETS]', concatenated_snippets)
            prompt = prompt.replace('[QUESTION]', question)
            cur_cost_args = CostReportingArgs(
                task_id=cost_args.task_id,
                user_id=cost_args.user_id,
                msg_id=cost_args.msg_id,
                description=cost_args.description + f" for corpus ID {corpus_id}",
                model=cost_args.model,
            )
            value_generation_params = {
                "user_prompt": prompt,
                "system_prompt": SYSTEM_PROMPT,
                "model": model,
                "fallback": GPT_4o,
                "response_format": PaperQAAnswer, 
            }
            output = llm_caller.call_method(
                cost_args=cur_cost_args,
                method=llm_completion,
                **value_generation_params,
            )
            # print(json.loads(output.result.content))
            response_simplified = {
                "question": question,
                "answer": json.loads(output.result.content)["answer"],
                "corpusId": corpus_id,
                "source": "vespa-snippets",
                "evidenceId": json.loads(output.result.content).get("exceprts", []),
                "cost": get_cost_object(output.result),
            }
            # print(response_simplified)
        else:
            response, cost = get_value_from_abstract(
                question=question, 
                corpus_id=corpus_id, 
                model=model,
                llm_caller=llm_caller,
                cost_args=cost_args,
            )
            response_simplified = {
                "question": question,
                "answer": response,
                "corpusId": corpus_id,
                "source": "abstract",
                "evidenceId": None,
                "cost": cost,
            }
    except Exception as e:
        logger.error(f"Exception while hitting vespa snippet search endpoint: {str(e)}")
        response_simplified = {"error": f"Exception while hitting vespa snippet search endpoint: {str(e)}"}
    return response_simplified


def generate_value_suggestions(
        column_name: str,
        column_def: str,
        corpus_ids: List[str],
        is_metadata: bool = False,
        model: str = "openai/gpt-4o-2024-08-06",
        paper_finder: PaperFinder = None,
        llm_caller: CostAwareLLMCaller = None,
        cost_args: CostReportingArgs = None,
) -> Dict:
    """
    Entry point to the cell value generation process.
    Given the name and definition of a column, we will
    produce values to be added to that column for each 
    paper in the table.
    """
    # Variables to track generated values, evidence and costs.
    cell_values = []
    evidence_ids = {}
    total_cost = 0.0

    # Setting #threads to parallelize value generation
    MAX_THREADS = 1
    # Setting snippeet search retrieval limit to 10 passages per paper
    paper_finder.retriever.n_retrieval = 10

    # Step 1: First, we check if the column to be populated is metadata-based.
    if is_metadata == "True":
        # If yes, we call the Semantic Scholar API to retrieve all metadata 
        # for each paper and construct a JSON blob containing this data.
        results = get_paper_metadata(corpus_ids)
        results = [results[x] if x in results else {} for x in corpus_ids]
        # We produce a query from the provided column name and definition.
        question = f"{column_name}, defined as {column_def}"
        raw_values = {}
        non_na_corpus_ids = []
    
        # We call our metadata-based value generation function with this query.
        # This is executed in parallel for all papers in the table for speed.
        # In addition to answers and costs, we also store corpus IDs for all papers
        # that have answers for the query (i.e., non-N/A values).
        with ThreadPoolExecutor(max_workers=MAX_THREADS) as executor:
            responses = list(executor.map(
                get_metadata_columns, 
                itertools.repeat(question), 
                results, 
                itertools.repeat(model),
                itertools.repeat(llm_caller),
                itertools.repeat(cost_args),
            ))
            raw_values = {y: x["answer"] if "answer" in x else "N/A" for x,y in zip(responses, corpus_ids)}
            non_na_corpus_ids = [x for x, y in raw_values.items() if y != 'N/A']
            per_cell_costs = {y: x.get("cost", None) for x,y in zip(responses, corpus_ids)}
    else:
        # For non-metadata column to be populated, we run value extraction 
        # on full-texts (backing off to abstracts) for all papers.
        # Step 1: Construct a query for value extraction using the column description.
        paperqa_query = f"Summarize this paper on the aspect of \"{column_name}\""
        if column_def != '':
            paperqa_query += f" by describing \"{column_def}\". "
            paperqa_query += "Only return the answer. Do not repeat the question or add any surrounding text. "
            paperqa_query += "The answer should be brief (fewer than 20 words) and need not be a complete sentence."
            paperqa_query += "Do not start the answer with references like \"this paper describes...\"."
        else:
            paperqa_query += f"."

        # Step 2: We call our QA-based value generation function with this query.
        # This is also executed in parallel for all papers, with storage of answers, evidence,
        # costs and corpus IDs for all papers that have answers for the query (i.e., non-N/A values).
        raw_values = {}
        non_na_corpus_ids = []
    
        with ThreadPoolExecutor(max_workers=MAX_THREADS) as executor:
            responses = list(executor.map(
                run_paper_qa, 
                itertools.repeat(paperqa_query), 
                corpus_ids, 
                itertools.repeat(model),
                itertools.repeat(paper_finder),
                itertools.repeat(llm_caller),
                itertools.repeat(cost_args),
            ))
            raw_values = {y: x["answer"] if "answer" in x else "No response" for x,y in zip(responses, corpus_ids)}
            non_na_corpus_ids = [x for x, y in raw_values.items() if y != 'N/A']
            evidence_ids = {y: x["evidenceId"] for x,y in zip(responses, corpus_ids) if "evidenceId" in x}
            per_cell_costs = {y: x.get("cost", None) for x,y in zip(responses, corpus_ids)}

    # Step 3: Construct final JSON blobs for each cell value containing answers
    # and evidence which can both be displayed on the UI.
    for k in non_na_corpus_ids:
        cell_value = {
            "corpusId": k,
            "displayValue": raw_values[k],
        }
        if evidence_ids and k in evidence_ids:
            cell_value['metadata'] = {
                "evidence": evidence_ids[k],
            }
        cell_values.append(cell_value)
    
    for k in corpus_ids:
        if  k not in non_na_corpus_ids:
            cell_value = {
            "corpusId": k,
            "displayValue": "N/A",
            }
            if evidence_ids and k in evidence_ids:
                cell_value['metadata'] = {
                    "evidence": evidence_ids[k],
                }
            cell_values.append(cell_value)

    return {"cell_values": cell_values, "cost": per_cell_costs}


import logging
from abc import abstractmethod
from typing import List, Dict, Any, Optional

import pandas as pd

from scholarqa.rag.reranker.reranker_base import AbstractReranker
from scholarqa.rag.retriever_base import AbstractRetriever
from scholarqa.utils import make_int, get_ref_author_str
from anyascii import anyascii

logger = logging.getLogger(__name__)


class AbsPaperFinder(AbstractRetriever):

    @abstractmethod
    def rerank(self, query: str, retrieved_ctxs: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        pass


class PaperFinder(AbsPaperFinder):
    snippet_srch_fields = ["text", "snippetKind", "snippetOffset", "section", "annotations.refMentions",
                           "annotations.sentences.start", "annotations.sentences.end"]

    def __init__(self, retriever: AbstractRetriever, context_threshold: float = 0.0, n_rerank: int = -1,
                 max_date: Optional[str] = None):
        self.retriever = retriever
        self.context_threshold = context_threshold
        self.n_rerank = n_rerank
        self.max_date = max_date

    def retrieve_passages(self, query: str, **filter_kwargs) -> List[Dict[str, Any]]:
        """Retrieve relevant passages along with scores from an index for the given query"""
        filter_kwargs.update({
            "fields": ",".join([f"snippet.{f}" for f in self.snippet_srch_fields])
        })
        if self.max_date:
            filter_kwargs.update({"insertedBefore": '{}'.format(self.max_date)})
        return self.retriever.retrieve_passages(query, **filter_kwargs)

    def retrieve_additional_papers(self, query: str, **filter_kwargs) -> List[Dict[str, Any]]:
        if self.max_date:
            # S2 API doesnt have insertedBefore for this endpoint, so use publicationDateOrYear:
            if year_filter := filter_kwargs.pop("year", None):
                # if year filter is present, we need to remove it from the kwargs and resolve it to publicationDateOrYear
                date_start, date_end = year_filter.split("-") # YYYY-YYYY
                max_year = self.max_date.split("-")[0] #YYYY-MM
                # restrict both start and end to max_year and if end is max_year, set it to max_date
                date_start, date_end = min(date_start, max_year), date_end if date_end and date_end < max_year else self.max_date
                filter_kwargs.update({"publicationDateOrYear": '{}:{}'.format(date_start, date_end)})

            else:
                filter_kwargs.update({"publicationDateOrYear": ':{}'.format(self.max_date)})

        return self.retriever.retrieve_additional_papers(query, **filter_kwargs)

    def rerank(self, query: str, retrieved_ctxs: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        return retrieved_ctxs

    def aggregate_into_dataframe(self, snippets_list: List[Dict[str, Any]], paper_metadata: Dict[str, Any]) -> \
            pd.DataFrame:
        """The reranked snippets is passage level. This function aggregates the passages to the paper level,
        The Dataframe also consists of aggregated passages stitched together with the paper title and abstract in the markdown format."""
        snippets_list = [snippet for snippet in snippets_list if snippet["corpus_id"] in paper_metadata
                         and snippet["text"] is not None]
        aggregated_candidates = self.aggregate_snippets_to_papers(snippets_list, paper_metadata)
        aggregated_candidates = [acand for acand in aggregated_candidates if
                                 acand["relevance_judgement"] >= self.context_threshold]

        return self.format_retrieval_response(aggregated_candidates)

    @staticmethod
    def aggregate_snippets_to_papers(snippets_list: List[Dict[str, Any]], paper_metadata: Dict[str, Any]) -> List[
        Dict[str, Any]]:
        logging.info(f"Aggregating {len(snippets_list)} passages at paper level with metadata")
        paper_snippets = dict()
        for snippet in snippets_list:
            corpus_id = snippet["corpus_id"]
            if corpus_id not in paper_snippets:
                paper_snippets[corpus_id] = paper_metadata[corpus_id]
                paper_snippets[corpus_id]["corpus_id"] = corpus_id
                paper_snippets[corpus_id]["sentences"] = []
            if snippet["stype"] != "public_api":
                paper_snippets[corpus_id]["sentences"].append(snippet)
            if "paperId" in paper_snippets[corpus_id]:
                del paper_snippets[corpus_id]["paperId"]
            paper_snippets[corpus_id]["relevance_judgement"] = max(
                paper_snippets[corpus_id].get("relevance_judgement", -1),
                snippet.get("rerank_score", snippet["score"]))
            if not paper_snippets[corpus_id]["abstract"] and snippet["section_title"] == "abstract":
                paper_snippets[corpus_id]["abstract"] = snippet["text"]
        sorted_ctxs = sorted(paper_snippets.values(), key=lambda x: x["relevance_judgement"], reverse=True)
        logger.info(f"Scores after aggregation: {[s['relevance_judgement'] for s in sorted_ctxs]}")
        return sorted_ctxs

    def format_retrieval_response(self, agg_reranked_candidates: List[Dict[str, Any]]) -> pd.DataFrame:
        def format_sections_to_markdown(row: List[Dict[str, Any]]) -> str:
            # convenience function to format the sections of a paper into markdown for function below
            # Convert the list of dictionaries to a DataFrame
            sentences_df = pd.DataFrame(row)
            if sentences_df.empty:
                return ""
            # Sort by 'char_offset' to ensure sentences are in the correct order
            sentences_df.sort_values(by="char_start_offset", inplace=True)

            # Group by 'section_title', concatenate sentences, and maintain overall order by the first 'char_offset'
            grouped = sentences_df.groupby("section_title", sort=False)["text"].apply("\n...\n".join)

            # Exclude sections titled 'Abstract' or 'Title'
            grouped = grouped[(grouped.index != "abstract") & (grouped.index != "title")]

            # Format as Markdown
            markdown_output = "\n\n".join(f"## {title}\n{text}" for title, text in grouped.items())
            return markdown_output

        df = pd.DataFrame(agg_reranked_candidates)
        try:
            df = df.drop(["text", "section_title", "ref_mentions", "score", "stype", "rerank_score"], axis=1)
        except Exception as e:
            logger.info(e)
        df = df[~df.sentences.isna() & ~df.year.isna()] if not df.empty else df
        if df.empty:
            return df
        df["corpus_id"] = df["corpus_id"].astype(int)

        # there are multiple relevance judgments in ['relevance_judgements'] for each paper
        # we will keep rows where ANY of the relevance judgments are 2 or 3
        # df = df[df["relevance_judgement"] >= self.context_threshold]

        if df.empty:
            return df

        # authors are lists of jsons. process with "name" key inside

        df["year"] = df["year"].apply(make_int)

        df["authors"] = df["authors"].fillna(value="")

        df.rename(
            columns={
                "citationCount": "citation_count",
                "referenceCount": "reference_count",
                "influentialCitationCount": "influential_citation_count",
            },
            inplace=True,
        )

        # drop corpusId
        df = df.drop(columns=["corpusId"])

        # now we need the big relevance_judgment_input_expanded
        # top of it
        # \n## Abstract\n{row['abstract']} --> Not using abstracts OR could use and not show
        prepend_text = df.apply(
            lambda
                row: f"# Title: {row['title']}\n# Venue: {row['venue']}\n"
                     f"# Authors: {', '.join([a['name'] for a in row['authors']])}\n## Abstract\n{row['abstract']}\n",
            axis=1,
        )
        section_text = df["sentences"].apply(format_sections_to_markdown)
        # update relevance_judgment_input
        df.loc[:, "relevance_judgment_input_expanded"] = prepend_text + section_text
        df["reference_string"] = df.apply(
            lambda
                row: anyascii(f"[{make_int(row.corpus_id)} | {get_ref_author_str(row.authors)} | "
                              f"{make_int(row['year'])} | Citations: {make_int(row['citation_count'])}]"),
            axis=1,
        )
        return df


class PaperFinderWithReranker(PaperFinder):
    def __init__(self, retriever: AbstractRetriever, reranker: AbstractReranker, n_rerank: int = -1,
                 context_threshold: float = 0.5, max_date: Optional[str] = None):
        super().__init__(retriever, context_threshold, n_rerank, max_date=max_date)
        if reranker:
            self.reranker_engine = reranker
        else:
            raise Exception(f"Reranker not initialized: {reranker}")

    def rerank(
            self, query: str, retrieved_ctxs: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Rerank the retrieved passages using a cross-encoder model and return the top n passages."""
        passages = [doc["title"] + " " + doc["text"] if "title" in doc else doc["text"] for doc in retrieved_ctxs]
        rerank_scores = self.reranker_engine.get_scores(
            query, passages
        )
        logger.info(f"Reranker scores: {rerank_scores}")

        for doc, rerank_score in zip(retrieved_ctxs, rerank_scores):
            doc["rerank_score"] = rerank_score
        sorted_ctxs = sorted(
            retrieved_ctxs, key=lambda x: x["rerank_score"], reverse=True
        )
        sorted_ctxs = super().rerank(query, sorted_ctxs)
        sorted_ctxs = sorted_ctxs[:self.n_rerank] if self.n_rerank > 0 else sorted_ctxs
        logging.info(f"Done reranking: {len(sorted_ctxs)} passages remain")
        return sorted_ctxs



from .scholar_qa import ScholarQA
from .rag.retrieval import PaperFinderWithReranker, PaperFinder
from .rag.retriever_base import FullTextRetriever, AbstractRetriever
from .rag.reranker.modal_engine import ModalReranker

__all__ = ["ScholarQA", "PaperFinderWithReranker", "PaperFinder", "FullTextRetriever", "AbstractRetriever",
           "ModalReranker", "llms", "postprocess", "preprocess",
           "utils", "models", "rag", "state_mgmt"]



import logging
import os
import sys
from collections import namedtuple
from logging import Formatter
from typing import Any, Dict, Optional, Set, List

import requests
from fastapi import HTTPException
from google.cloud import storage

from scholarqa import glog
from scholarqa.llms.litellm_helper import setup_llm_cache

logger = logging.getLogger(__name__)

S2_APIKEY = os.getenv("S2_API_KEY", "")
S2_HEADERS = {"x-api-key": S2_APIKEY}
S2_API_BASE_URL = "https://api.semanticscholar.org/graph/v1/"
NUMERIC_META_FIELDS = {"year", "citationCount", "referenceCount", "influentialCitationCount"}
CATEGORICAL_META_FIELDS = {"title", "abstract", "corpusId", "authors", "venue", "isOpenAccess", "openAccessPdf"}
METADATA_FIELDS = ",".join(CATEGORICAL_META_FIELDS.union(NUMERIC_META_FIELDS))


class TaskIdAwareLogFormatter(Formatter):
    def __init__(self, task_id: str = ""):
        super().__init__("%(asctime)s - %(name)s - %(levelname)s")
        self.task_id = task_id

    def format(self, record):
        og_message = super().format(record)
        task_id_part = f"[{self.task_id}] " if self.task_id else ""
        return f"{og_message} - {task_id_part}- {record.getMessage()}"


def init_settings(logs_dir: str, log_level: str = "INFO",
                  litellm_cache_dir: str = "litellm_cache") -> TaskIdAwareLogFormatter:
    def setup_logging() -> TaskIdAwareLogFormatter:
        # If LOG_FORMAT is "google:json" emit log message as JSON in a format Google Cloud can parse
        loggers = [
            "LiteLLM Proxy",
            "LiteLLM Router",
            "LiteLLM"
        ]

        for logger_name in loggers:
            litellm_logger = logging.getLogger(logger_name)
            litellm_logger.setLevel(logging.WARNING)

        fmt = os.getenv("LOG_FORMAT")
        tid_log_fmt = TaskIdAwareLogFormatter()
        if fmt == "google:json":
            handlers = [glog.Handler()]
            for handler in handlers:
                handler.setFormatter(glog.Formatter(tid_log_fmt))
        else:
            handlers = []
            # log lower levels to stdout
            stdout_handler = logging.StreamHandler(stream=sys.stdout)
            stdout_handler.addFilter(lambda rec: rec.levelno <= logging.INFO)
            handlers.append(stdout_handler)

            # log higher levels to stderr (red)
            stderr_handler = logging.StreamHandler(stream=sys.stderr)
            stderr_handler.addFilter(lambda rec: rec.levelno > logging.INFO)
            handlers.append(stderr_handler)
            for handler in handlers:
                handler.setFormatter(tid_log_fmt)

        level = log_level
        logging.basicConfig(level=level, handlers=handlers)
        return tid_log_fmt

    def setup_local_llm_cache():
        # Local logs directory for litellm caching, event traces and state management
        local_cache_dir = f"{logs_dir}/{litellm_cache_dir}"
        # create parent and subdirectories for the local cache
        os.makedirs(local_cache_dir, exist_ok=True)
        setup_llm_cache(cache_type="disk", disk_cache_dir=local_cache_dir)

    tid_log_fmt = setup_logging()
    setup_local_llm_cache()
    return tid_log_fmt


def make_int(x: Optional[Any]) -> int:
    try:
        return int(x)
    except:
        return 0


def get_ref_author_str(authors: List[Dict[str, str]]) -> str:
    if not authors:
        return "NULL"
    f_author_lname = authors[0]["name"].split()[-1]
    return f_author_lname if len(authors) == 1 else f"{f_author_lname} et al."


def query_s2_api(
        end_pt: str = "paper/batch",
        params: Dict[str, Any] = None,
        payload: Dict[str, Any] = None,
        method="get",
):
    url = S2_API_BASE_URL + end_pt
    req_method = requests.get if method == "get" else requests.post
    response = req_method(url, headers=S2_HEADERS, params=params, json=payload)
    if response.status_code != 200:
        logging.exception(f"S2 API request to end point {end_pt} failed with status code {response.status_code}")
        raise HTTPException(
            status_code=500,
            detail=f"S2 API request failed with status code {response.status_code}",
        )
    return response.json()


def get_paper_metadata(corpus_ids: Set[str], fields=METADATA_FIELDS) -> Dict[str, Any]:
    if not corpus_ids:
        return {}
    paper_data = []
    corpus_ids = list(corpus_ids)
    for idx in range(0, len(corpus_ids), 100):
        logger.info(f"Processing {idx} - {idx+100}/{len(corpus_ids)}")
        batch_ids = corpus_ids[idx: idx + 100]
        batch_data = query_s2_api(
            end_pt="paper/batch",
            params={
                "fields": fields
            },
            payload={"ids": ["CorpusId:{0}".format(cid) for cid in batch_ids]},
            method="post",
        )
        paper_data.extend(batch_data)

    paper_metadata = {
        str(pdata["corpusId"]): {k: make_int(v) if k in NUMERIC_META_FIELDS else pdata.get(k) for k, v in pdata.items()}
        for pdata in paper_data if pdata and "corpusId" in pdata
    }
    return paper_metadata


def push_to_gcs(text: str, bucket: str, file_path: str):
    try:
        storage_client = storage.Client()
        bucket = storage_client.bucket(bucket)
        blob = bucket.blob(file_path)
        blob.upload_from_string(text)
        logging.info(f"Pushed event trace: {file_path} to GCS")
    except Exception as e:
        logging.info(f"Error pushing {file_path} to GCS: {e}")



# Data structure to represent a table
from pydantic import BaseModel, Field
from typing import Any, Dict, List, Optional

class TableWidget(BaseModel):
    id: str = Field(description="Unique identifier for the table widget")
    rows: List["TableRow"] = Field(
        default_factory=list, description="List of rows in the table"
    )
    columns: List["TableColumn"] = Field(
        default_factory=list, description="List of columns in the table"
    )
    cells: Dict[str, "TableCell"] = Field(
        default_factory=dict,
        description="Dictionary of cells in the table. Key is in the format ${row.id}-${column.id}",
    )
    title: Optional[str] = Field(default=None, description="Title for table generated")

    def add_rows(self, new_rows: List["TableRow"]) -> None:
        self.rows.extend(new_rows)

    def add_columns(self, new_columns: List["TableColumn"]) -> None:
        self.columns.extend(new_columns)
    
    def to_dict(self) -> dict:
        table = {}
        table["id"] = self.id
        table["title"] = self.title
        table["rows"] = [row.__dict__ for row in self.rows]
        table["columns"] = [column.__dict__ for column in self.columns]
        table["cells"] = {k: v.__dict__ for k, v in self.cells.items()}
        return table


class TableRow(BaseModel):
    id: str = Field(description="Unique identifier for the row")
    display_value: Optional[str] = Field(
        default=None, description="Display value of the row"
    )
    paper_corpus_id: Optional[int] = Field(description="Corpus id of a paper")


class TableColumn(BaseModel):
    id: str = Field(description="Unique identifier for the column")
    name: str = Field(description="Name of the column")
    description: str = Field(description="Description of the column")
    is_metadata: Optional[bool] = Field(
        default=False,
        description="Flag indicating whether the column to be populated is a metadata column or not",
    )
    tools: List[str] = Field(
        description="List of tools used to generate the column values"
    )


class TableCell(BaseModel):
    id: str = Field(
        description="Unique identifier for the cell. In the format ${row.id}-${column.id}"
    )
    value: Optional[str | int] = Field(default=None, description="Value of the cell")
    display_value: str = Field(description="Display value of the cell")
    is_loading: bool = Field(
        default=False, description="Whether the cell is still loading"
    )
    error: Optional[str] = Field(default=None, description="Error message if any")
    metadata: Optional[Dict[str, Any]] = Field(
        default=None, description="Metadata of the cell"
    )
    edited_value: Optional[str] = Field(
        default=None,
        description="overridden cell value by a user, nora, or other 3rd party",
    )
    edited_by_uuid: Optional[str] = Field(
        default=None, description="Who last edited the table cell"
    )


from pythonjsonlogger import jsonlogger

import logging


class Formatter(jsonlogger.JsonFormatter):
    """
    Custom log formatter that emits log messages as JSON, with the "severity" field
    which Google Cloud uses to differentiate message levels.
    """

    def __init__(self, task_id_aware_formatter = None):
        super().__init__()
        self.task_id_aware_formatter = task_id_aware_formatter

    def add_fields(self, log_record, record, message_dict):
        super().add_fields(log_record, record, message_dict)
        log_record["severity"] = record.levelname

    def format(self, record: logging.LogRecord) -> str:
        record.msg = {"message": self.task_id_aware_formatter.format(record)} if self.task_id_aware_formatter else record.msg
        return super().format(record)



class Handler(logging.StreamHandler):
    def __init__(self, stream=None):
        super().__init__(stream)
        self.setFormatter(Formatter())



from abc import ABC, abstractmethod
from google.cloud import storage
import logging
import json
import os

logger = logging.getLogger(__name__)


class TraceWriter(ABC):
    @abstractmethod
    def write(self, trace_json, file_name: str) -> None:
        pass


class GCSWriter(TraceWriter):
    def __init__(self, bucket_name: str):
        self.bucket_name = bucket_name

    def write(self, trace_json, file_name: str) -> None:
        try:
            trace_json_str = json.dumps(trace_json.__dict__)
            storage_client = storage.Client()
            bucket = storage_client.bucket(self.bucket_name)
            blob = bucket.blob(f"{file_name}.json")
            blob.upload_from_string(trace_json_str)
            logger.info(f"Pushed event trace: {file_name}.json to GCS")
        except Exception as e:
            logger.info(f"Error pushing {file_name} to GCS: {e}")


class LocalWriter(TraceWriter):
    def __init__(self, local_dir: str):
        self.local_dir = local_dir
        if not os.path.exists(local_dir):
            logger.info(f"Creating local directory to record traces: {local_dir}")
            os.makedirs(local_dir)

    def write(self, trace_json, file_name: str) -> None:
        try:
            with open(f"{self.local_dir}/{file_name}.json", "w") as f:
                json.dump(trace_json.__dict__, f, indent=4)
            logger.info(f"Pushed event trace to local path: {self.local_dir}/{file_name}.json")
        except Exception as e:
            logger.info(f"Error pushing {file_name} to local directory: {e}")



from typing import List, Optional, Any, Dict

from nora_lib.tasks.models import AsyncTaskState as BaseAsyncTaskState
from pydantic import BaseModel, Field
from scholarqa.table_generation.table_model import TableWidget


class Author(BaseModel):
    name: str = Field(description="The name of the author")
    authorId: Optional[str] = Field(description="The Semantic Scholar id of the author")


class PaperDetails(BaseModel):
    corpus_id: int = Field(allow_none=False, description=(
        "The Semantic Scholar id of the cited paper"))
    title: str = Field(description="Title of the paper")
    year: int = Field(description="Year of publication")
    venue: Optional[str] = Field(description="Venue of publication", default=None)
    authors: List[Author] = Field(description="Authors of the paper", default=None)
    n_citations: Optional[int] = Field(default=0, description=(
        "The number of times the source paper has been cited"
    ))


# TODO: define your request data
class ToolRequest(BaseModel):
    task_id: Optional[str] = Field(default=None, description=(
        "Reference to a long-running task. Provide this argument to receive an update on its"
        "status and possibly its result."
    ))
    query: str = Field(default=None, description=(
        "A scientific query posed to scholar qa by a user"
    ))
    opt_in: Optional[bool] = Field(default=True, description=(
        "Flag to indicate whether to include the query and response in public release"))
    user_id: Optional[str] = Field(default=None, description="The user id of the user who posed the query")


class CitationSrc(BaseModel):
    id: str = Field(default=None, description=(
        "The id of the citation which is of the format (index, author_ref_string, year)"
    ))
    paper: PaperDetails = Field(description="Metadata of the cited paper")
    snippets: Optional[List[str]] = Field(default=[], description=(
        "A list of all the relevant snippets from the cited paper"
    ))
    score: float = Field(description=("Relevance score of the snippet for the query"))


class GeneratedSection(BaseModel):
    title: str = Field(default=None, description=(
        "header for the generated section text"
    ))
    tldr: str = Field(default=None, description=(
        "A short summary of the generated section"
    ))
    text: str = Field(default=None, description=(
        "The generated section text"
    ))
    citations: List[CitationSrc] = Field(default=None, description=(
        "The citations used in the generated section"
    ))
    table: Optional[TableWidget] = Field(default=None, description=("Table widget object for sections with list format"))


# TODO: define your result data
class TaskResult(BaseModel):
    """The outcome of running a Task to completion"""
    sections: List[GeneratedSection] = Field(
        description="The generated iterations of the answer"
    )
    cost: float = Field(description="The overall cost of the task", default=0.0)
    tokens: Optional[Dict[str, int]] = Field(description="The tokens used to generate the answer", default=None)


class ToolResponse(BaseModel):
    task_id: str = Field(description="Unique identifiers for the invocation of the tool.")
    query: str = Field(description="The query that was posed to the tool.")
    task_result: TaskResult


class TaskStep(BaseModel):
    description: str = Field(description="The step in the task")
    start_timestamp: float = Field(description="The timestamp when the step was started")
    estimated_timestamp: float = Field(description="The estimated timestamp for the step to complete", default=None)


class AsyncTaskState(BaseAsyncTaskState[TaskResult]):
    pass


class AsyncToolResponse(BaseModel):
    task_id: str = Field(
        "Identifies the long-running task so that its status and eventual result"
        "can be checked in follow-up calls."
    )
    query: str = Field(description="The query that was posed to the tool.")
    estimated_time: str = Field(description="How long we expect this task to take from start to finish")
    task_status: str = Field(description="Current human-readable status of the task.")
    task_result: Optional[TaskResult] = Field(description="Final result of the task.")
    steps: List[TaskStep] = Field(description="The steps processed in the task so far.", default=[])


if __name__ == "__main__":
    import json

    # task_result_schema = TaskResult.model_json_schema()
    # with open("task_result.schema.json", "w") as f:
    #     json.dump(task_result_schema, f, indent=2)

    async_task_state_schema = AsyncTaskState.model_json_schema()
    with open("async_task_state.schema.json", "w") as f:
        json.dump(async_task_state_schema, f, indent=2)



import logging
import json
from pydantic import BaseModel
from typing import List, Dict

from scholarqa.table_generation.prompts import ATTRIBUTE_PROMPT, SYSTEM_PROMPT
from scholarqa.utils import get_paper_metadata
from scholarqa.llms.litellm_helper import CostAwareLLMCaller, CostReportingArgs, llm_completion
from scholarqa.llms.constants import GPT_4o

logger = logging.getLogger(__name__)

# Definition of output format to be provided during json mode
class Column(BaseModel):
    name: str
    definition: str
    is_metadata: bool

class ColumnSuggestions(BaseModel):
    columns: List[Column]

def retrieve_paper_info(corpus_ids: List[str]) -> Dict:
    """ 
    Given a set of corpus IDs for papers to be added to the table,
    retrieve titles and abstracts for each paper using the
    Semantic Scholar batch querying API.
    """
    paper_metadata = get_paper_metadata(corpus_ids)
    paper_info = {corpus_id: paper_metadata[corpus_id] if corpus_id in paper_metadata else {} for corpus_id in corpus_ids}
    return paper_info


def format_paper_info(paper_info: Dict) -> str:
    """
    Given titles and abstracts of all papers in the table,
    format this information to be appended to the column suggestion prompt.
    """
    formatted_paper_info = ""
    for index, corpus_id in enumerate(paper_info):
        paper = paper_info[corpus_id]
        title = paper["title"] if "title" in paper else None
        abstract = paper["abstract"].strip() if "abstract" in paper and paper["abstract"] else None
        formatted_paper_info += f'Paper {index+1} title: {title}\nPaper {index+1} abstract: {abstract}\n\n'
    return formatted_paper_info


def generate_final_prompt(query: str, formatted_paper_info: str, column_num: str) -> str:
    """
    Given the formatted paper information, and an optional user query,
    generate the final column suggestion prompt to be sent to the LLM.
    """
    final_prompt = ATTRIBUTE_PROMPT.format(query, column_num, formatted_paper_info)
    return final_prompt


def generate_attribute_suggestions(
        corpus_ids: List[str], 
        model: str = GPT_4o, 
        query: str = None,
        column_num: int = 10,
        llm_caller: CostAwareLLMCaller = None,
        cost_args: CostReportingArgs = None,
    ) -> Dict:
    """
    Entry point to the column suggestion generation process.
    """
    # Step 1: Retrieve user query or backoff to the default query
    default_user_query = "Brief Overview and Comparison of Following Papers"
    user_query = query if query is not None else default_user_query

    # Step 2: Retrieve titles and abstracts for all provided papers
    paper_info = retrieve_paper_info(corpus_ids)
    
    # Step 3: Format all paper titles and abstracts to add to prompt
    formatted_paper_info = format_paper_info(paper_info)
    
    # Step 4: Produce final column generation prompt from papers and user query
    final_prompt = generate_final_prompt(user_query, formatted_paper_info, column_num)

    # Step 5: Prompt the LLM to produce column suggestions
    column_suggestion_params = {
        "user_prompt": final_prompt,
        "system_prompt": SYSTEM_PROMPT,
        "response_format": ColumnSuggestions,
        "model": model,
    }
    output = llm_caller.call_method(
        cost_args=cost_args,
        method=llm_completion,
        **column_suggestion_params,
    )
    column_suggestions = json.loads(output.result.content)["columns"]
    cost_dict = {
        "cost_value": output.result.cost,
        "tokens": {
            "total": output.result.total_tokens,
            "prompt": output.result.input_tokens,
            "completion": output.result.output_tokens,
            "reasoning": output.result.reasoning_tokens
        },
        "model": output.result.model,
    }

    return {"columns": column_suggestions, "cost": cost_dict}


import json
import logging
import re
from enum import Enum
from typing import Tuple, Dict, List, Any, Generator

import pandas as pd
from pydantic import BaseModel, Field

from scholarqa.llms.constants import GPT_4o
from scholarqa.llms.litellm_helper import batch_llm_completion, llm_completion
from scholarqa.llms.prompts import USER_PROMPT_PAPER_LIST_FORMAT, USER_PROMPT_QUOTE_LIST_FORMAT, \
    PROMPT_ASSEMBLE_NO_QUOTES_SUMMARY
from scholarqa.llms.constants import CompletionResult

logger = logging.getLogger(__name__)


class DimFormat(str, Enum):
    SYNTHESIS = "synthesis"
    LIST = "list"


class Dimension(BaseModel):
    name: str = Field(default=None, description=(
        "The name of the dimension"
    ))
    format: DimFormat = Field(default=None, description=(
        "The generation format of the dimension - can be either list of synthesis"
    ))
    quotes: List[int] = Field(default=None, description=(
        "A list of indices of paper quotes in the dimension, can be empty if no relevant quotes are found"
    ))


class ClusterPlan(BaseModel):
    cot: str = Field(default=None, description=(
        "The justification for every dimension name and its format"
    ))
    dimensions: List[Dimension] = Field(default=None, description=(
        "The list of dimensions along with the associated quote indices as per the cot plan"
    ))


class MultiStepQAPipeline:
    def __init__(self, llm_model: str, fallback_llm: str = GPT_4o, batch_workers: int = 20,
                 **llm_kwargs):
        self.llm_model = llm_model
        self.fallback_llm = fallback_llm
        self.batch_workers = batch_workers
        self.llm_kwargs = {"max_tokens": 4096*4}
        if llm_kwargs:
            self.llm_kwargs.update(llm_kwargs)


    def step_select_quotes(self, query: str, scored_df: pd.DataFrame, sys_prompt: str) -> Tuple[
        Dict[str, str], List[CompletionResult]]:

        logger.info(
            f"Querying {self.llm_model} to extract quotes from these papers with {self.batch_workers} parallel workers")
        tup_items = {k: v for k, v in
                     zip(scored_df["reference_string"], scored_df["relevance_judgment_input_expanded"])}
        messages = [USER_PROMPT_PAPER_LIST_FORMAT.format(query, v) for k, v in tup_items.items()]
        completion_results = batch_llm_completion(self.llm_model, messages=messages, system_prompt=sys_prompt,
                                                  max_workers=self.batch_workers, fallback=self.fallback_llm,
                                                  **self.llm_kwargs)
        quotes = [
            cr.content if cr.content != "None" and not cr.content.startswith("None\n") and not cr.content.startswith(
                "None ")
            else "" for cr in completion_results]
        per_paper_summaries = {t[0]: quote for t, quote in zip(tup_items.items(), quotes) if len(quote) > 10}
        per_paper_summaries = dict(sorted(per_paper_summaries.items(), key=lambda x: x[0]))
        return per_paper_summaries, completion_results

    def step_clustering(self, query: str, per_paper_summaries: Dict[str, str],
                        sys_prompt: str) -> Tuple[Dict[str, Any], CompletionResult]:
        def make_prompt(query: str, paper_paper_quotes_dict: Dict[str, str]) -> str:
            # paper_paper_quotes_dict is a dictionary with keys being the paper titles and values being the quotes
            # need to make a single string with all of the quotes
            quotes = ""
            for idx, (paper, quotes_str) in enumerate(paper_paper_quotes_dict.items()):
                # there are multiple quotes per paper
                quotes_str = quotes_str.replace("\n", "")
                quotes += f"[{idx}]\t{quotes_str}" + "\n"
            prompt = USER_PROMPT_QUOTE_LIST_FORMAT.format(query, quotes)
            return prompt

        user_prompt = make_prompt(query, per_paper_summaries)
        try:
            #params for reasoning mode: max_completion_tokens=4096, max_tokens=4096+1024, reasoning_effort="low"
            response = llm_completion(user_prompt=user_prompt,
                                      system_prompt=sys_prompt, fallback=self.fallback_llm, model=self.llm_model,
                                      response_format= ClusterPlan, **self.llm_kwargs
                                      )
            return json.loads(response.content), response
        except Exception as e:
            logger.warning(f"Error while clustering with {self.llm_model}: {e}")
            raise e

    def generate_iterative_summary(self, query: str, per_paper_summaries_extd: Dict[str, Dict[str, Any]],
                                   plan: Dict[str, Any],
                                   sys_prompt: str) -> Generator[CompletionResult, None, None]:
        # first, we need to make a map from the index to the quotes because the llm is using index only

        # now fill in the prompt
        per_paper_summaries_tuples = [(ref_string, response) for ref_string, response in
                                      per_paper_summaries_extd.items()]
        # only use the section headings from the plan, discard the quote indices
        plan_str = "\n".join([k for k in plan])
        existing_sections = []
        i = 0
        for section_name, inds in plan.items():
            # inds are a string like this: "[1, 2, 3]"
            # get the quotes for each index
            quotes = ""
            for ind in inds:
                if ind < len(per_paper_summaries_tuples):
                    quotes += (
                            per_paper_summaries_tuples[ind][0] + ": " + str(per_paper_summaries_tuples[ind][
                                                                                1]) + "\n"
                    )
                else:
                    logger.warning(f"index {ind} out of bounds")
            # existing sections should have their summaries removed because they are confusing.
            # remove anything in []
            already_written = "\n\n".join(existing_sections)
            already_written = re.sub(r"\[.*?\]", "", already_written)
            fill_in_prompt_args = {
                "query": query,
                "plan": plan_str,
                "already_written": already_written,
                "section_name": section_name}
            if quotes:
                fill_in_prompt_args["section_references"] = quotes
                filled_in_prompt = sys_prompt.format(**fill_in_prompt_args)
            else:
                logger.warning(f"No quotes for section {section_name}")
                filled_in_prompt = PROMPT_ASSEMBLE_NO_QUOTES_SUMMARY.format(**fill_in_prompt_args)

            response = llm_completion(user_prompt=filled_in_prompt, model=self.llm_model, fallback=self.fallback_llm,
                                      **self.llm_kwargs)
            existing_sections.append(response.content)
            yield response



# step 1 prompt
SYSTEM_PROMPT_QUOTE_PER_PAPER = """
In this task, you are presented with a user query and an academic paper with snippets and metadata.

Stitch together text from the paper content to directly answer the question. 

To be clear, copy EXACT text ONLY.

Include any references that are part of the text to be copied. The references can occur at the beginning, middle, or end of the text.

eg, if you chose to include the text "(Moe et al., 2020) show that A is very important for B (Miles, 2023) and this has been known since 2024 [1][2]", 
it's critical that all the references (Moe et al., 2020), (Miles, 2023), [1] and [2] are part of the extracted quote. Include all forms of academic citation if they are contiguous with your selected quote.

Use ... to indicate that there is a gap of excluded text between text you chose.

For example: Text to answer... More text here... start a sentence in the middle.

No need to use the title. 

Sometimes you will see authors and/or section titles. Do not use them in your answer.

Output the quote ONLY. Do not introduce it with any text, formatting, or white spaces.

If the paper does not answer the user query at all, just output None
"""

USER_PROMPT_PAPER_LIST_FORMAT = """
Here is the user's query:<user_query>
{}
</user_query>
And here is the paper with snippets and metadata that may have salient content for the query:
<paper_with_snippets>
{}
</paper_with_snippets>"""

# step 2 prompts
CLUSTER_PROMPT_FEW_SHOTS = """For example, if the user query is "Is true that: Language models are not universally better at discriminating among previously generated alternatives than generating initial responses."
Then the DIMENSIONS could be "language models studied", "discrimination approaches", "discrimination performance", etc

Another example, if the user query is "What are alternatives to the ADAM algorithm?" 
Then you can either have a single dimension: "Alternatives to ADAM" or you can have multiple dimensions like "Alternatives to ADAM in Optimization", "Alternatives to ADAM in Deep Learning". 
"""

CLUSTER_PROMPT_DIRECTIVE = """For each section, decide if it should be a bullet-point list or a synthesis paragraph. 
Bullet-point lists are right when the user wants a list or table of items. 
Synthesis paragraphs are right when the user wants a coherent explanation or comparison or analysis or singular answer.

For user queries that are simpler, you only need one dimension. 
For example, if the query is "I want a list of alternatives to GPT-3" or "Give me a table of GPT-3 alternatives", then there is only dimension "List of Alternatives to GPT-3".
"""

SYSTEM_PROMPT_QUOTE_CLUSTER = f"""
In this task, you are presented with quoted passages that were collected from a set of papers.
The goal is to fuse all those quotes into a summary answer/report that satisfies the user query in an easy-to-consume format.

As an intermediate step, please cluster quotes for different section in the summary. 
First, output a set of DIMENSIONS that break down the user query for a scientific audience. 

{CLUSTER_PROMPT_FEW_SHOTS}

You should start by making a plan of which candidate dimensions make sense to clearly and directly answer the query, ignoring the snippets.

The first section should be title "Introduction" or "Background" to provide the user the key basics needed to understand the rest of the answer.

{CLUSTER_PROMPT_DIRECTIVE}

IMPORTANT: Make sure the clusters are in the same order you would then write the corresponding summary.
IMPORTANT: Make sure that EVERY input quote is included somewhere in the output.
IMPORTANT: Some sections may not have any quotes to support them. Include them in the output JSON regardless with an empty list. This is particularly true of the "Introduction" or "Background" section.

Choose list or synthesis with deep care and wisdom. Start with a markdown justification for the section name and its format.

The last thing you output is an assignment of each quote to a dimension like so:  
{{
"cot": "Justification for every dimension name and its format...",
"dimensions": [{{"name": "dimension name 1", "format": "synthesis or list", "quotes": [comma-delimited highlights indices]}},
{{"name": "dimension name 2", "format": "synthesis or list", "quotes": [comma-delimited highlights indices]}},
{{"name": "dimension name 3", "format": "synthesis or list", "quotes": []}},  # empty because we didn't find any supporting quotes
...
]
}}
Dont add unnecessary linebreaks or spaces to the output.
"""

USER_PROMPT_QUERY_FORMAT = """
Here is the user's query:
    <user_query>
    {}
    </user_query>
"""

USER_PROMPT_QUOTE_LIST_FORMAT = USER_PROMPT_QUERY_FORMAT + """And here are the quotes from the papers that may have salient content for the query:
    <quotes>
    {}
    </quotes>"""

# step 3 prompt
PROMPT_ASSEMBLE_SUMMARY = """
A user issued a query and a set of papers were provided with salient content.
The user query was: {query}

Here is the overall plan for the summary answer:

<plan>
{plan}
</plan>

I will provide you with the name of one section from the plan at a time, along with the list of chosen quotes for that section. 
Your job is to help me write this section and cite the provided quoted references. 

Here is what has already been written:
<already_written_sections>
{already_written}
</already_written_sections>

The section I would like you to write next is (type of section):
<section_name>
{section_name}
</section_name>

Here are the reference quotes to cite for this section:
<section_references>
{section_references}
</section_references>

<citation instructions>
- Each reference is a key value pair, where the key is a pipe separated string enclosed in square brackets representing [ID | AUTHOR_REF | YEAR | Citations: CITES].

The value consists of the quote and sometimes a dictionary of inline citations referenced in that quote
eg. "[2345677 | Doe, Moe et al. | 2024 | Citations: 25]": {{"quote": "This is the reference text.", "inline citations": {{"[4517277 | Hero et al. | 2019 | Citations: 250]": "This is an inline citation."}}}}

- Please write this section, making sure to cite the relevant references inline using the corresponding reference key in the format: [ID | AUTHOR_REF | YEAR | Citations: CITES]. You may use more than one reference key in a row if it's appropriate. In general, use all of the references that support your written text.

- Along with the quote, if any of its accompanying inline citations are relevant to or mentioned in the claim you are writing, you should cite them too using the same aforementioned format. 

For example, let's say you write 

"The X was shown to be Y." [1234 | A | 2023 | Citations: 3]. 

And the reference A (2023) you want to cite states "As shown in B (2020), X is Y." In this case, reference 1234 is textually direct support for what you wrote, but it itself clearly states that B (2020) is the actual source of this information. As such, you need to cite both, one after the other as 

"The X was shown to be Y." [1234 | A | 2023 | Citations: 3] [4321 | B | 202 | Citations: 25]

- You can add something from your own knowledge. This should only be done if you are sure about its truth and/or if there is not enough information in the references to answer the user's question. Cite the text from your knowledge as [LLM MEMORY | 2024]. The citation should follow AFTER the text. Don't cite LLM Memory with another evidence source. 

- Note that all citations that support what you write must come after the test you write. That's how humans read in-line cited text. First text, then the citation.
</citation instructions>

<writing instructions>
The section should have the following characteristics:
- Before the section write a 2 sentence "TLDR;" of the section. No citations here. Precede with the text "TLDR;"
- Use direct and simple language everywhere, like "use" and "can". Avoid using more complex words if simple ones will do.
- Use the citation count to decide what is "notable" or "important". If the citation count is 100 or more, you are allowed to use value judgments like "notable."
- Some references are older. Something that claims to be "state of the art" but is from 2020 may not be any more. Please avoid making such claims.
- Be concise.
- The section you write must be a coherent continuation of already_written and directly answer the user query.
- The section you write should add to the user's understanding of what they learned in already_written. 
- Multiple references may express the same idea. If so, you can cite multiple references in a single sentence.
- Do not make the same points that were made in the previous already written sections.
</writing instructions>

<format and structure instructions>
Start the section with its section_name and then a newline and then the text "TLDR;", the actual TLDR, and then write the summary. 
Rules for section formatting:
- For example, if the section name in the plan is "Important Papers (list)", then write it as "Important Papers" and format the section as a LIST (required). 
- For example, if the section name in the plan is "Deep Dive on Networks (synthesis)" then render it as "Deep Dive on Networks" and write a SYNTHESIS paragraph (required).
- The section format MUST match what's in the parentheses of the section name. A list HAS to be a list. a SYNTHESIS has to be a paragraph. Seriously.
- Write the section content using markdown format
</format and structure instructions>
"""

PROMPT_ASSEMBLE_NO_QUOTES_SUMMARY = """
A user issued a query and the goal is to provide an answer based on the given below plan.
The user query was: {query}

Here is the overall plan for the summary answer:

<plan>
{plan}
</plan>

I will provide you with the name of one section from the plan at a time. 
Your job is to help me write this section. 

Here is what has already been written:
<already_written_sections>
{already_written}
</already_written_sections>

The section I would like you to write next is (type of section):
<section_name>
{section_name}
</section_name>


<citation instructions>
- Please write this section. This should only be done if you are sure about its truth.
- Cite the text as [LLM MEMORY | 2024]. The citation should follow AFTER the text.
</citation instructions>

<writing instructions>
The section should have the following characteristics:
- Before the section write a 2 sentence "TLDR;" of the section. No citations here. Precede with the text "TLDR;"
- Use direct and simple language everywhere, like "use" and "can". Avoid using more complex words if simple ones will do.
- Be concise.
- The section you write must be a coherent continuation of already_written and directly answer the user query.
- The section you write should add to the user's understanding of what they learned in already_written. 
- Multiple references may express the same idea. If so, you can cite multiple references in a single sentence.
- Do not make the same points that were made in the previous already written sections.
- Fix weird unicode.
</writing instructions>

<format and structure instructions>
Start the section with its section_name and then a newline and then the text "TLDR;", the actual TLDR, and then expand upon it. 
Rules for section formatting:
- For example, if the section name in the plan is "Important Papers (list)", then write it as "Important Papers" and format the section as a LIST (required). 
- For example, if the section name in the plan is "Deep Dive on Networks (synthesis)" then render it as "Deep Dive on Networks" and write a SYNTHESIS paragraph (required).
- The section format MUST match what's in the parentheses of the section name. A list HAS to be a list. a SYNTHESIS has to be a paragraph. Seriously.
</format and structure instructions>
"""

QUERY_DECOMPOSER_PROMPT = """
<task>
Your task is to analyze a query issued by a user of an academic question-answering system and break it down into parts relevant for searching and retrieving high-quality, well-cited answers. The goal is to create a structured JSON output that can be used by an academic search engine API, which supports filters like publication years, venues, authors, and fields of study.

Your output should decompose the query into the following components:
1. Publication years: If the query specifies a time range (e.g., “recent” or “last five years”), convert it to the relevant year range. If no time range is mentioned, leave these fields blank.  
   - The current year is 2025. Interpret “recent” as 2022–2025 and adjust other relative terms (e.g., “last decade,” “since 2018”) accordingly.
2. Venues: Include any journals, conferences, or publishers mentioned explicitly in the query as a comma separated string. Use their exact names.
3. Authors: List any authors mentioned explicitly in the query. Each author name should appear as a separate entry in an array.
4. Fields of study: Use only the following fields of study. If the query includes subfields or ambiguous terms, map them to the closest match from this list:  
   - Computer Science, Medicine, Chemistry, Biology, Materials Science, Physics, Geology, Psychology, Art, History, Geography, Sociology, Business, Political Science, Economics, Philosophy, Mathematics, Engineering, Environmental Science, Agricultural and Food Sciences, Education, Law, Linguistics.  
   Concatenate multiple fields with commas and no spaces (e.g., `Physics,Mathematics`).
5. Rewritten query: Simplify the remaining query into a concise, natural-language phrase, excluding any information already extracted into `year`, `venues`, `authors`, or `fields_of_study`.
6. Rewritten query for keyword search: Remove unnecessary stop words and connectors to create a keyword-friendly version of the remaining query, excluding any information already extracted into other fields.
7. Complex, multi-sentence queries should still be complete in terms of content when rewritten. The goal of the rewritten query is to remove the metadata like year, venues, authors and fields_of_study, but keep all of the important topical content that needs to be addressed in an answer.

<note about handling ambiguous terms and missing information>  
- If a field cannot be inferred from the query (e.g., no authors are mentioned), leave it empty.  
- For terms not matching the fields of study list, map them to the closest matching field(s). For instance, “machine learning” should map to `Computer Science`, while “neuroscience” might map to `Biology` or `Psychology` based on context.  
</note about handling ambiguous terms and missing information>  
</task>

<examples>

<example input #1>  
What are the latest papers by Andrew Ng on deep reinforcement learning?  
</example input #1>

<example output #1>  
```json
{
    "earliest_search_year": "2022",
    "latest_search_year": "2025",
    "venues": "",
    "authors": ["Andrew Ng"],
    "field_of_study": "Computer Science",
    "rewritten_query": "Deep reinforcement learning.",
    "rewritten_query_for_keyword_search": "deep reinforcement learning"
}
```
</example output #1>  

<example input #2>  
Summarize the findings on climate policy impacts in articles published in Nature or Science from the last five years.  
</example input #2>

<example output #2>  
```json
{
    "earliest_search_year": "2020",
    "latest_search_year": "2025",
    "venues": "Nature,Science",
    "authors": [],
    "field_of_study": "Environmental Science,Political Science",
    "rewritten_query": "Findings on climate policy impacts.",
    "rewritten_query_for_keyword_search": "climate policy impacts"
}
```
</example output #2>  

<example input #3>  
Discuss recent contributions by Noam Chomsky to the study of linguistics and cognitive science.  
</example input #3>

<example output #3>  
```json
{
    "earliest_search_year": "2022",
    "latest_search_year": "2025",
    "venues": "",
    "authors": ["Noam Chomsky"],
    "field_of_study": "Linguistics,Psychology",
    "rewritten_query": "Contributions to linguistics and cognitive science.",
    "rewritten_query_for_keyword_search": "linguistics cognitive science"
}
```
</example output #3>  

<example input #4>  
What are the effects of climate change on agricultural productivity in Sub-Saharan Africa in recent studies?
</example input #4>

<example output #4>  
```json
{
    "earliest_search_year": "2022",
    "latest_search_year": "2025",
    "venues": "",
    "authors": [],
    "field_of_study": "Environmental Science,Agricultural and Food Sciences",
    "rewritten_query": "Effects of climate change on agricultural productivity in Sub-Saharan Africa.",
    "rewritten_query_for_keyword_search": "climate change agricultural productivity Sub-Saharan Africa"
}
```
</example output #4>  

<example input #5>  
Explore the role of neural networks in solving mathematical optimization problems. 
</example input #5>

<example output #5>  
```json
{
    "earliest_search_year": "",
    "latest_search_year": "",
    "venues": "",
    "authors": [],
    "field_of_study": "Computer Science,Mathematics",
    "rewritten_query": "Role of neural networks in solving mathematical optimization problems.",
    "rewritten_query_for_keyword_search": "neural networks mathematical optimization problems"
}
```
</example output #5>  

<example input #6>  
Discuss the historical significance of the Renaissance period on modern art and philosophy.
</example input #6>

<example output #6>  
```json
{
    "earliest_search_year": "",
    "latest_search_year": "",
    "venues": "",
    "authors": [],
    "field_of_study": "Art,Philosophy",
    "rewritten_query": "Historical significance of Renaissance on modern art and philosophy.",
    "rewritten_query_for_keyword_search": "Renaissance modern art philosophy"
}
```
</example output #6>  

<example input #7>  
Review papers by Andrew Ng and Yann LeCun on neural networks since 2010. 
</example input #7>

<example output #7>  
```json
{
    "earliest_search_year": "2010",
    "latest_search_year": "2025",
    "venues": "",
    "authors": ["Andrew Ng", "Yann LeCun"],
    "field_of_study": "Computer Science",
    "rewritten_query": "Neural networks.",
    "rewritten_query_for_keyword_search": "neural networks"
}
```
</example output #7> 
</examples>
"""



import os
from abc import ABC, abstractmethod
from time import time
from typing import List, Any, Optional, Union, Tuple
from uuid import uuid5, UUID

from nora_lib.tasks.state import IStateManager, StateManager

from scholarqa.llms.constants import CompletionResult, CostReportingArgs, TokenUsage
from scholarqa.models import TaskResult, TaskStep, AsyncTaskState, ToolRequest

UUID_NAMESPACE = os.getenv("UUID_ENCODER_KEY", "ai2-scholar-qa")


class AbsStateMgrClient(ABC):
    @abstractmethod
    def get_state_mgr(self, tool_req: ToolRequest) -> IStateManager:
        pass

    def init_task(self, task_id: str, tool_request: ToolRequest):
        pass

    def update_task_state(
            self,
            task_id: str,
            tool_req: ToolRequest,
            status: str,
            step_estimated_time: int = 0,
            curr_response: Any = None,
            task_estimated_time: str = None,
    ):
        state_mgr = self.get_state_mgr(tool_req)
        curr_step = TaskStep(description=status, start_timestamp=time())
        task_state = state_mgr.read_state(task_id)
        task_state.task_status = status
        if step_estimated_time:
            curr_step.estimated_timestamp = curr_step.start_timestamp + step_estimated_time
        if task_estimated_time:
            task_state.estimated_time = task_estimated_time
        if curr_response:
            task_state.task_result = TaskResult(sections=curr_response)
        task_state.extra_state["steps"].append(curr_step)
        state_mgr.write_state(task_state)

    def report_llm_usage(self, completion_costs: List[CompletionResult], cost_args: CostReportingArgs) -> float:
        pass


class LocalStateMgrClient(AbsStateMgrClient):
    def __init__(self, logs_dir: str, async_state_dir: str = "async_state"):
        self._async_state_dir = f"{logs_dir}/{async_state_dir}"
        os.makedirs(self._async_state_dir, exist_ok=True)
        self.state_mgr = StateManager(AsyncTaskState, self._async_state_dir)

    def get_state_mgr(self, tool_req: Optional[ToolRequest] = None) -> IStateManager:
        return self.state_mgr

    def report_llm_usage(self, completion_costs: List[CompletionResult], cost_args: CostReportingArgs) -> Union[float, Tuple[float, TokenUsage]]:
        tot_cost = sum([cost.cost for cost in completion_costs])
        token_usage = TokenUsage(
            input=sum([cost.input_tokens for cost in completion_costs]),
            output=sum([cost.output_tokens for cost in completion_costs]),
            total=sum([cost.total_tokens for cost in completion_costs]),
            reasoning=sum([cost.reasoning_tokens for cost in completion_costs])
        )
        return tot_cost, token_usage

    def init_task(self, task_id: str, tool_request: ToolRequest):
        try:
            tool_request.user_id = str(uuid5(namespace=UUID(tool_request.user_id), name=f"nora-{UUID_NAMESPACE}"))
        except Exception as e:
            pass



from typing import Dict, Any, Optional, Tuple, Union, List

import modal
import os

from scholarqa.rag.reranker.reranker_base import AbstractReranker, RERANKER_MAPPING
import logging

logger = logging.getLogger(__name__)


class ModalReranker(AbstractReranker):
    def __init__(self, app_name: str, api_name: str, batch_size=32, gen_options: Dict[str, Any] = None):
        logger.info(f"using model {app_name} on Modal for reranking")
        self.modal_engine = ModalEngine(app_name, api_name, gen_options)
        self.batch_size = batch_size

    def get_scores(self, query: str, documents: List[str]):
        logger.info("Invoking the reranker deployed on Modal")
        return self.modal_engine.generate(
            (query, documents, self.batch_size), streaming=False
        )


RERANKER_MAPPING["modal"] = ModalReranker


class ModalEngine:
    modal_client: modal.Client

    def __init__(self, model_id: str, api_name: str, gen_options: Dict[str, Any] = None) -> None:
        # Skiff secrets
        modal_token = os.getenv("MODAL_TOKEN")
        modal_token_secret = os.getenv("MODAL_TOKEN_SECRET")
        self.modal_client = modal.Client.from_credentials(modal_token, modal_token_secret)
        self.model_id = model_id
        self.gen_options = {"max_tokens": 1024, "temperature": 0.7, "logprobs": 2,
                            "stop_token_ids": [128009]} if gen_options is None else gen_options
        self.api_name = api_name

    def fn_lookup(
            self,
            **opt_kwargs,
    ) -> Tuple[modal.Function, Optional[Dict[str, Any]]]:
        if opt_kwargs:
            opts = {**self.gen_options, **opt_kwargs} if self.gen_options else {**opt_kwargs}
        else:
            opts = self.gen_options

        fn = modal.Function.from_name(self.model_id, self.api_name)
        fn.hydrate(client=self.modal_client)
        return fn, opts if opts else None

    def generate(self, input_args: Tuple, streaming=False, **opt_kwargs) -> Union[str, List[Dict]]:
        gen_fn, opts = self.fn_lookup(**opt_kwargs)
        if streaming:
            outputs = []
            if opts:
                for chunk in gen_fn.remote_gen(*input_args, opts):
                    outputs.append(chunk)
            else:
                for chunk in gen_fn.remote_gen(*input_args):
                    outputs.append(chunk)
            return "".join(outputs) if outputs and type(outputs[0]) == str else outputs
        else:
            return gen_fn.remote(*input_args, **opts) if opts else gen_fn.remote(*input_args)



from collections import namedtuple

GPT_4_TURBO = "openai/gpt-4-turbo-2024-04-09"
GPT_4o = "openai/gpt-4o-2024-08-06"
GPT_4o_MINI = "openai/gpt-4o-mini"
GPT_4_1 = "openai/gpt-4.1"
GPT_5_CHAT= "openai/gpt-5-chat-latest"
CLAUDE_3_OPUS = "anthropic/claude-3-opus-20240229"
CLAUDE_35_SONNET = "anthropic/claude-3-5-sonnet-20241022"
CLAUDE_37_SONNET = "anthropic/claude-3-7-sonnet-20250219"
LLAMA_405_TOGETHER_AI = "together_ai/meta-llama/Meta-Llama-3.1-405B-Instruct-Turbo"
CLAUDE_4_SONNET = "anthropic/claude-sonnet-4-20250514"

CompletionResult = namedtuple("CompletionCost",
                              ["content", "model", "cost", "input_tokens", "output_tokens", "total_tokens", "reasoning_tokens"])

CostReportingArgs = namedtuple("CostReportingArgs", ["task_id", "user_id", "msg_id", "description", "model"])

TokenUsage = namedtuple("TokenUsage", ["input", "output", "total", "reasoning"])

CostAwareLLMResult = namedtuple("CostAwareLLMResult", ["result", "tot_cost", "models", "tokens"])



import logging
import os
import re
from threading import Thread
from time import time
from typing import List, Any, Dict, Tuple, Generator
from uuid import uuid4

import pandas as pd
from anyascii import anyascii
from langsmith import traceable

from scholarqa.config.config_setup import LogsConfig
from scholarqa.llms.constants import CostAwareLLMResult, CLAUDE_4_SONNET, CLAUDE_37_SONNET, GPT_5_CHAT
from scholarqa.llms.litellm_helper import CostAwareLLMCaller, CostReportingArgs
from scholarqa.llms.prompts import SYSTEM_PROMPT_QUOTE_PER_PAPER, SYSTEM_PROMPT_QUOTE_CLUSTER, PROMPT_ASSEMBLE_SUMMARY
from scholarqa.models import GeneratedSection, TaskResult, ToolRequest, CitationSrc
from scholarqa.postprocess.json_output_utils import get_json_summary
from scholarqa.preprocess.query_preprocessor import validate, decompose_query, LLMProcessedQuery
from scholarqa.rag.multi_step_qa_pipeline import MultiStepQAPipeline
from scholarqa.rag.retrieval import PaperFinder
from scholarqa.state_mgmt.local_state_mgr import AbsStateMgrClient, LocalStateMgrClient
from scholarqa.trace.event_traces import EventTrace
from scholarqa.utils import get_paper_metadata, NUMERIC_META_FIELDS, CATEGORICAL_META_FIELDS, get_ref_author_str, \
    make_int
from scholarqa.table_generation.table_model import TableWidget
from scholarqa.table_generation.table_generator import TableGenerator

logger = logging.getLogger(__name__)

# Regular expressions to fix weird formatting issues cause after citation linking in the evidences
CLOSE_BRACKET_PATTERN = r'(?<![\[|,\s*\d])(\d+\])'  # (Doe et al., 2024)10] --> (Doe et al., 2024)[10]
OPEN_BRACKET_PATTERN = r"(\[[\d+,]+),(?=[^\[]*$)"  # [8,9,(Doe et al., 2024) --> [8,9](Doe et al., 2024)


class ScholarQA:
    def __init__(
            self,
            # Required for webapp since a new process is created for each request, for library task_id can be None initially and assigned for each request as below
            paper_finder: PaperFinder,
            task_id: str = None,
            llm_model: str = CLAUDE_4_SONNET,
            multi_step_pipeline: MultiStepQAPipeline = None,
            state_mgr: AbsStateMgrClient = None,
            logs_config: LogsConfig = None,
            run_table_generation: bool = True,
            llm_kwargs: Dict[str, Any] = None,
            **kwargs
    ):
        if logs_config:
            self.logs_config = logs_config
        else:
            logger.info("initializing the log configs")
            self.logs_config = LogsConfig(llm_cache_dir="lib_llm_cache")
            self.logs_config.init_formatter()

        self.task_id = task_id
        self.paper_finder = paper_finder
        self.llm_model = llm_model
        fallback_llm = kwargs.get("fallback_llm", f"{GPT_5_CHAT},{CLAUDE_37_SONNET}")
        self.validate = kwargs.get("validate", "OPENAI_API_KEY" in os.environ)
        if not self.validate:
            logger.warning("Validation of the query for harmful content is turned off")
        self.decomposer_llm = kwargs.get("decomposer_llm", self.llm_model)
        self.state_mgr = state_mgr if state_mgr else LocalStateMgrClient(self.logs_config.log_dir)
        self.llm_caller = CostAwareLLMCaller(self.state_mgr)
        self.llm_kwargs = llm_kwargs if llm_kwargs else dict()
        if not multi_step_pipeline:
            logger.info(f"Creating a new MultiStepQAPipeline with model: {llm_model} for all the steps")
            self.multi_step_pipeline = MultiStepQAPipeline(self.llm_model, fallback_llm=fallback_llm, **self.llm_kwargs)
        else:
            self.multi_step_pipeline = multi_step_pipeline

        self.tool_request = None
        self.table_llm = kwargs.get("table_llm", self.llm_model)
        self.table_generator = TableGenerator(paper_finder=paper_finder, llm_caller=self.llm_caller)
        self.run_table_generation = run_table_generation

    def update_task_state(
            self,
            status: str,
            step_estimated_time: int = 0,
            curr_response: List[GeneratedSection] = None,
            task_estimated_time: str = None,
    ):
        logger.info(status)
        if self.task_id and self.tool_request:
            self.state_mgr.update_task_state(
                self.task_id,
                self.tool_request,
                status,
                step_estimated_time,
                curr_response,
                task_estimated_time,
            )

    @traceable(name="Preprocessing: Validate and decompose user query")
    def preprocess_query(self, query: str, cost_args: CostReportingArgs=None, ) -> CostAwareLLMResult:
        if self.validate:
            # Validate the query for harmful/unanswerable content
            validate(query)
        # Decompose the query to get filters like year, venue, fos, citations, etc along with a re-written
        # version of the query and a query suitable for keyword search.
        llm_args = {"max_tokens": 4096*2}
        if self.llm_kwargs:
            llm_args.update(self.llm_kwargs)
        return self.llm_caller.call_method(
            cost_args=cost_args, method=decompose_query, query=query, decomposer_llm_model=self.decomposer_llm,
            fallback=self.multi_step_pipeline.fallback_llm, **llm_args
        )

    @traceable(name="Retrieval: Find relevant paper passages for the query")
    def find_relevant_papers(self, llm_processed_query: LLMProcessedQuery, **kwargs) -> Tuple[
        List[Dict[str, Any]], List[Dict[str, Any]]]:
        # retrieval from vespa index
        start = time()
        rewritten_query = llm_processed_query.rewritten_query
        keyword_query = llm_processed_query.keyword_query
        self.update_task_state(
            f"Retrieving relevant passages from a corpus of 8M+ open access papers",
            step_estimated_time=5
        )
        # Get relevant paper passages from the Semantic Scholar index for the llm rewritten query
        snippet_results = self.paper_finder.retrieve_passages(query=rewritten_query,
                                                              **llm_processed_query.search_filters,
                                                              **kwargs)
        snippet_corpus_ids = {snippet["corpus_id"] for snippet in snippet_results}
        self.update_task_state(f"Retrieved {len(snippet_results)} highly relevant passages", step_estimated_time=1)

        if keyword_query:
            # Get additional papers from the Semantic Scholar api via keyword search
            search_api_results = self.paper_finder.retrieve_additional_papers(keyword_query,
                                                                              **llm_processed_query.search_filters)
            search_api_results = [item for item in search_api_results if item["corpus_id"] not in snippet_corpus_ids]
            self.update_task_state(
                f"Retrieved {len(search_api_results)} more papers from Semantic Scholar abstracts using keyword search",
                step_estimated_time=1)
        else:
            search_api_results = []
        logger.info("Retrieval time: %.2f", time() - start)

        return snippet_results, search_api_results

    @traceable(name="Retrieval: Rerank the passages and aggregate at paper level")
    def rerank_and_aggregate(self, user_query: str, retrieved_candidates: List[Dict[str, Any]], filter_paper_metadata: [
        Dict[str, Any]]) -> Tuple[pd.DataFrame, Dict[str, Any]]:
        if self.paper_finder.n_rerank > 0:
            self.update_task_state(
                f"Further re-rank and aggregate passages to focus on up to top {self.paper_finder.n_rerank} papers",
                step_estimated_time=10)
        start = time()
        reranked_candidates = self.paper_finder.rerank(user_query, retrieved_candidates)
        logger.info("Reranking time: %.2f", time() - start)
        paper_metadata = filter_paper_metadata
        paper_metadata.update(get_paper_metadata(
            {snippet["corpus_id"] for snippet in reranked_candidates if
             snippet["corpus_id"] not in filter_paper_metadata}))
        agg_df = self.paper_finder.aggregate_into_dataframe(reranked_candidates, paper_metadata)
        self.update_task_state(
            f"Found {len(agg_df)} highly relevant papers after re-ranking and aggregating",
            step_estimated_time=1)
        logger.info("Reranking w. formatting time: %.2f", time() - start)
        return agg_df, paper_metadata

    @traceable(name="Generation: Extract relevant quotes from paper passages or filter")
    def step_select_quotes(self, query: str, scored_df: pd.DataFrame, cost_args: CostReportingArgs = None,
                           sys_prompt: str = SYSTEM_PROMPT_QUOTE_PER_PAPER) -> CostAwareLLMResult:
        logger.info("Running Step 1 - quote extraction")
        self.update_task_state("Extracting salient key statements from papers",
                               step_estimated_time=15)
        logger.info(
            f"{scored_df.shape[0]} papers with relevance_judgement >= {self.paper_finder.context_threshold} to start with.")
        start = time()
        cost_args = cost_args._replace(model=self.multi_step_pipeline.llm_model)._replace(
            description="Corpus QA Step 1: Quote extraction")
        per_paper_summaries = self.llm_caller.call_method(cost_args, self.multi_step_pipeline.step_select_quotes,
                                                          query=query, scored_df=scored_df,
                                                          sys_prompt=sys_prompt)
        api_corpus_ids = set(
            scored_df[scored_df.sentences.apply(lambda x: not x)].corpus_id.astype(str))
        ref_strs = {rs.split(" | ")[0][1:] for rs in per_paper_summaries.result}
        logger.info(f"Paper abstracts used from s2 api: {api_corpus_ids.intersection(ref_strs)}")

        logger.info(
            f"Step 1 done - {len(per_paper_summaries.result)} papers with quotes extracted, cost: {per_paper_summaries.tot_cost}, "
            f"time: {time() - start:.2f}")
        return per_paper_summaries

    @traceable(name="Generation: Cluster quotes to generate an organization plan")
    def step_clustering(self, query: str, per_paper_summaries: Dict[str, str], cost_args: CostReportingArgs = None,
                        sys_prompt: str = SYSTEM_PROMPT_QUOTE_CLUSTER) -> CostAwareLLMResult:
        logger.info("Running Step 2: Clustering the extracted quotes into meaningful dimensions")
        self.update_task_state("Synthesizing an answer outline based on extracted quotes", step_estimated_time=15)
        start = time()
        cost_args = cost_args._replace(model=self.multi_step_pipeline.llm_model)._replace(
            description="Corpus QA Step 2: Clustering quotes into dimensions")
        cluster_json = self.llm_caller.call_method(cost_args, self.multi_step_pipeline.step_clustering,
                                                   query=query, per_paper_summaries=per_paper_summaries,
                                                   sys_prompt=sys_prompt)
        logger.info(f"Step 2 done - {cluster_json.result}, cost: {cluster_json.tot_cost}, time: {time() - start:.2f}")
        return cluster_json

    @traceable(name="Generation: Generate an iterative summary")
    def step_gen_iterative_summary(self, query: str, per_paper_summaries: Dict[str, str],
                                   plan_json: Dict[str, Any], cost_args: CostReportingArgs = None,
                                   sys_prompt: str = PROMPT_ASSEMBLE_SUMMARY) -> Generator[
        str, None, CostAwareLLMResult]:
        logger.info("Running Step 3: Assemble the summary with the links (takes ~2 mins)")
        start = time()

        cost_args = cost_args._replace(model=self.multi_step_pipeline.llm_model)._replace(
            description="Corpus QA Step 3: Generating summarized answer")
        sec_generator = self.llm_caller.call_iter_method(cost_args, self.multi_step_pipeline.generate_iterative_summary,
                                                         query=query, per_paper_summaries_extd=per_paper_summaries,
                                                         plan=plan_json, sys_prompt=sys_prompt)
        try:
            while True:
                response = next(sec_generator)
                yield response.content
        except StopIteration as e:
            return_val = e.value
        if return_val:
            logger.info(f"Step 3 done, cost: {return_val.tot_cost}, time: {time() - start:.2f}")
        return return_val

    @staticmethod
    def passage_to_quotes_metadata(retrieval_df: pd.DataFrame, per_paper_summaries: Dict[str, str],
                                   plan_json: Dict[str, List[int]]) -> Dict[str, List[Dict[str, Any]]]:
        """
        Map the quotes extracted in step 1 `per_paper_summaries` to their actual full length versions
        in retrieval_df and make a map of quote ref string --> [inline citations].

        i) First parse the plan json to get the required paper indices to be sent for the final generation in step 3.
        ii) For those papers, find the quotes from per_paper_summaries and passages from retrieval_df.
        iii) For each quote, first check if the paper metadata from retrieval_df consists of passages, if not, then the quotes are taken from
        abstracts of keyword results.
        iii) Else, try to match the paper quotes as substrings to the corresponding paper passages to get the inline citations along with
        offsets.
        iv) First try matching the raw string (high precision), if unsuccessful, try matching only the alphabet characters (high recall).
        v) Once a match is found, use its offsets in the passage and iterate over the sentence offsets/inline citations to find any that occur within the snippet.
        vi) In case of a raw string match, replace any citation mentions (if possible), with the paper id of the corresponding inline citation to be linked later.
        """
        # get all the ref strings for the clutering plan generated in step 2
        ref_str_list = [k for k in per_paper_summaries]
        # paper identifiers for the selected quotes in the plan obtained from their corresponding index
        req_ref_strs = {ref_str_list[item] for sublist in plan_json.values() for item in sublist if
                        item < len(ref_str_list)}
        quotes_metadata = dict()
        if req_ref_strs:
            # filter the quotes according to the plan
            reqd_paper_summaries = {k: v for k, v in per_paper_summaries.items() if k in req_ref_strs}
            # filter the dataframe according to the plan
            reqd_ref_df = retrieval_df[retrieval_df["reference_string"].apply(lambda x: x in req_ref_strs)].copy()
            # remove all special characters from the passages in the dataframe for approximate match
            reqd_ref_df["sentence_alpha"] = reqd_ref_df["sentences"].apply(
                lambda x: [re.sub(r'[^a-zA-Z]', '', sentence["text"]).lower() for sentence in x])
            # iterate over the reqd_ref_df and get the snippets for each row from reqd_paper_summaries
            for row_idx, row in reqd_ref_df.iterrows():
                ref_str, sentences, sent_alpha = row["reference_string"], row["sentences"], row["sentence_alpha"]
                mapped_quotes = []

                curr_reqd_quotes = reqd_paper_summaries[ref_str].split("...")
                curr_reqd_quotes_reg = [re.sub(r'[^a-zA-Z]', '', quote).lower() for quote in curr_reqd_quotes]
                for idx, (quote, quote_reg) in enumerate(zip(curr_reqd_quotes, curr_reqd_quotes_reg)):
                    new_quote = quote.strip()
                    curr_quote_map = {"quote": new_quote, "section_title": "abstract",
                                      "pdf_hash": "", } if not sentences else dict()

                    shift = 0  # keep track of changes to the quote offsets when the inline citations are modified
                    for sidx, sentence in enumerate(sentences):
                        # can lookup exact string now since we prompt the llm to include the citations in the quotes
                        lookup_idx = sentence["text"].lower().find(quote.lower().strip())
                        raw_match = lookup_idx >= 0
                        if not raw_match:
                            lookup_idx = sent_alpha[sidx].find(quote_reg)
                        if lookup_idx >= 0:
                            lookup_end = lookup_idx + len(quote)
                            curr_quote_map["section_title"] = sentence["section_title"]
                            curr_quote_map["pdf_hash"] = sentence["pdf_hash"]
                            curr_quote_map["start"], curr_quote_map["end"] = lookup_idx, lookup_end
                            curr_quote_map["sentence_offsets"], curr_quote_map["ref_mentions"] = [], []
                            if sentence.get("sentence_offsets"):
                                for sidx, soff in enumerate(sentence["sentence_offsets"]):
                                    # check if the sentence offset is within the range of the quote
                                    # the sentence can be completely or partially inside the quote
                                    if (lookup_idx < soff["end"] <= lookup_end) or (
                                            lookup_idx <= soff["start"] < lookup_end)\
                                            or (soff["start"] <= lookup_idx and lookup_end <= soff["end"]):
                                        curr_quote_map["sentence_offsets"].append(soff)
                            if sentence.get("ref_mentions"):
                                for sref in sentence["ref_mentions"]:
                                    if sref.get("matchedPaperCorpusId") and lookup_idx <= sref.get(
                                            "start") and sref.get("end") <= lookup_end:
                                        curr_quote_map["ref_mentions"].append(sref["matchedPaperCorpusId"])
                                        if raw_match:
                                            new_start, new_end = sref["start"] - lookup_idx, sref["end"] - lookup_idx
                                            cite_str = f"({sref['matchedPaperCorpusId']})"
                                            new_quote = new_quote[:new_start + shift] + cite_str + new_quote[
                                                                                                   new_end + shift:]
                                            shift += (len(cite_str) - sref["end"] + sref["start"])
                                # curr_inline_citations.update(
                                #     [sref["matchedPaperCorpusId"] for sref in sentence["ref_mentions"] if
                                #      sref.get("start") >= lookup_idx and sref.get("end") <= lookup_end])
                            break
                    curr_quote_map["quote"] = new_quote
                    if "section_title" not in curr_quote_map:
                        curr_quote_map["pdf_hash"] = ""
                        for field in ["title", "abstract"]:
                            if row[field] and new_quote.lower() in row[field].lower():
                                curr_quote_map["section_title"] = field
                    mapped_quotes.append(curr_quote_map)
                quotes_metadata[ref_str] = mapped_quotes
                updated_quotes = "...".join([mq["quote"] for mq in mapped_quotes])
                # fix weird formatting
                updated_quotes = re.sub(CLOSE_BRACKET_PATTERN, r'[\1',
                                        updated_quotes)  # (Doe et al., 2024)10] --> (Doe et al., 2024)[10]
                updated_quotes = re.sub(OPEN_BRACKET_PATTERN, r'\1]',
                                        updated_quotes)  # [8,9,(Doe et al., 2024) --> [8,9](Doe et al., 2024)
                per_paper_summaries[ref_str] = updated_quotes

        return quotes_metadata

    def populate_citations_metadata(self, avl_paper_metadata: Dict[str, Dict[str, Any]],
                                    paper_inline_cites: Dict[str, List],
                                    per_paper_summaries: Dict[str, Any]) -> Dict[str, Dict[str, Any]]:
        """
        retrieve the metadata of the quote inline citations if not already present and update the quotes from string,
        to a dict of {"quote": quote, "inline_citations": {ref_str: abstract, ... }}.
        Also, link the citation mentions to modify them from (corpus_id) to (Doe et al., 2024)
        """
        corpus_id_ref_str_map = {ref_str[1:-1].split(" | ")[0]: ref_str for ref_str in per_paper_summaries}
        additional_citation_ids = {item for sublist in paper_inline_cites.values() for item in sublist if
                                   item not in avl_paper_metadata}
        if additional_citation_ids:
            logger.info(f"Fetching metadata for {len(additional_citation_ids)} additional inline citations")
            additional_metadata = get_paper_metadata(additional_citation_ids)
        else:
            additional_metadata = dict()
        per_paper_summaries = {k: {"quote": quote, "inline_citations": dict()} for k, quote in
                               per_paper_summaries.items()}
        for ref_str, cite_ids in paper_inline_cites.items():
            quote_cid = ref_str[1:-1].split(" | ")[0]
            # 2 sets of ref mentions, one where paper metadata is already available and the rest for which metadata was requested from s2 api
            curr_metadata = [avl_paper_metadata[cite_id] for cite_id in cite_ids if
                             cite_id in avl_paper_metadata]
            curr_metadata += [additional_metadata[cite_id] for cite_id in cite_ids if
                              cite_id in additional_metadata]
            for idx, mdata in enumerate(curr_metadata):
                mref_str = corpus_id_ref_str_map.get(mdata["corpusId"], f"[{mdata['corpusId']} | "
                                                                        f"{get_ref_author_str(mdata['authors'])} | "
                                                                        f"{make_int(mdata.get('year'))} "
                                                                        f"| Citations: {make_int(mdata['citationCount'])}]")
                mref_str = anyascii(mref_str)
                per_paper_summaries[ref_str]["quote"] = per_paper_summaries[ref_str]["quote"].replace(
                    f"({mdata['corpusId']})",
                    f"({get_ref_author_str(mdata['authors'])}, {make_int(mdata.get('year'))})")
                if mdata["corpusId"] in additional_metadata:
                    additional_metadata[mdata["corpusId"]]["relevance_judgement"] = max(
                        additional_metadata[mdata["corpusId"]].get("relevance_judgement", 0),
                        avl_paper_metadata[quote_cid]["relevance_judgement"])
                if mdata.get("abstract"):
                    per_paper_summaries[ref_str]["inline_citations"][mref_str] = mdata["abstract"]
            per_paper_summaries[ref_str]["quote"] = per_paper_summaries[ref_str]["quote"].replace("NULL, ", "")
        avl_paper_metadata.update(additional_metadata)
        return per_paper_summaries

    def extract_quote_citations(self, score_df: pd.DataFrame, per_paper_summaries: Dict[str, str],
                                plan_json: Dict[str, List[int]], paper_metadata: Dict[str, Any]) -> Tuple[
        Dict[str, Dict[str, Any]], Dict[str, List[Dict[str, Any]]]]:
        quotes_metadata = self.passage_to_quotes_metadata(score_df, per_paper_summaries, plan_json)
        per_paper_inline_cites = {
            ref_str: set(ref for q in qmeta if q.get("ref_mentions") for ref in q["ref_mentions"])
            for ref_str, qmeta in quotes_metadata.items()
        }
        per_paper_inline_cites = {k: sorted(v) for k, v in per_paper_inline_cites.items() if v}
        per_paper_summaries_extd = self.populate_citations_metadata(paper_metadata, per_paper_inline_cites,
                                                                    per_paper_summaries)
        for ref_str, quote_map in per_paper_summaries_extd.items():
            if quotes_metadata.get(ref_str):
                quote_parts = quote_map["quote"].split("...")
                for idx, qmap in enumerate(quotes_metadata[ref_str]):
                    qmap["quote"] = quote_parts[idx]
        return per_paper_summaries_extd, quotes_metadata

    @staticmethod
    def get_gen_sections_from_json(section: Dict[str, Any]) -> GeneratedSection:
        try:
            citations = [CitationSrc(**citation) for citation in section["citations"]]
            generated_section = GeneratedSection(title=section["title"],
                                                 tldr=section["tldr"],
                                                 text=section["text"],
                                                 citations=citations)
            return generated_section
        except Exception as e:
            logger.error(f"Error while converting json to TaskResult: {e}")
            raise e

    def postprocess_json_output(self, json_summary: List[Dict[str, Any]], **kwargs) -> None:
        pass

    def answer_query(self, query: str, inline_tags: bool = True) -> Dict[str, Any]:
        task_id = str(uuid4())
        self.logs_config.task_id = task_id
        logger.info("New task")
        tool_request = ToolRequest(task_id=task_id, query=query, user_id="lib_user")
        try:
            task_result = self.run_qa_pipeline(tool_request, inline_tags)
        except Exception as e:
            logger.warning(f"Error while running task: {e}, invalidating llm cache and retrying")
            self.multi_step_pipeline.llm_kwargs["cache"] = {"no-cache": True}
            task_result = self.run_qa_pipeline(tool_request, inline_tags)
        return task_result.model_dump()

    def gen_table_thread(self, user_id: str, query: str, dim: Dict[str, Any],
                         cit_ids: List[int], tlist: List[Any]) -> Thread:
        def call_table_generator(didx: int, payload: Dict[str, Any]):
            logger.info(
                "Received table generation request for topic: " + payload["section_title"]
            )
            table, costs = self.table_generator.run_table_generation(
                thread_id=payload["task_id"],
                user_id=payload["user_id"],
                original_query=payload["query"],
                section_title=payload["section_title"],
                corpus_ids=payload["cit_ids"],
                column_model=payload["column_model"],
                value_model=payload["value_model"],
            )
            tlist[dim["idx"]] = (table, costs)
            
        task_id = self.task_id if self.task_id else self.tool_request.task_id
        payload = {
            "task_id": task_id,
            "user_id": user_id,
            "query": query,
            "section_title": dim["name"],
            "cit_ids": cit_ids,
            "column_model": self.table_llm,
            "value_model": self.table_llm,
        }
        tthread = Thread(target=call_table_generator, args=(dim["idx"], payload,))
        tthread.start()
        return tthread

    def get_user_msg_id(self):
        return self.tool_request.user_id, self.task_id

    @traceable(run_type="tool", name="ai2_scholar_qa_trace")
    def run_qa_pipeline(self, req: ToolRequest, inline_tags=False) -> TaskResult:
        """
                This function takes a query and returns a response.
                Goes through the following steps:
                0) Decompose the query to get filters like year, venue, fos, citations, etc along with a re-written
                version of the query and a query suitable for keyword search.
                1) Query retrieval to get the relevant snippets from the index (n_retrieval)
                1.1) Query semantic scholar with the keyword search query to get the relevant papers.(n_keyword_srch)
                2) Re-rank the snippets based on the query with a cross encoder (n_rerank)
                3) Get exact relevant quotes from an LLM
                4) Generate outline and cluster the quotes from (3)
                4.1) The quotes cluster in the outline have inline citations associated with them. Map the quotes to
                their inline citations and include them with the quotes.
                5) Generate the summarized output using the quotes and outline in (3) and (4)

                :param req: A scientific query posed to scholar qa by a user, consists of the string query, task id and user id
                :param inline_tags: Whether to include inline <paper> tags in the output or not
                :return: A response to the query

        """
        self.tool_request = req
        self.update_task_state("Processing user query", task_estimated_time="~3 minutes", step_estimated_time=5)
        task_id = self.task_id if self.task_id else req.task_id
        user_id, msg_id = self.get_user_msg_id()
        msg_id = task_id if not msg_id else msg_id
        query = req.query
        logger.info(
            f"Received query: {query} from user_id: {user_id} with opt_in: {req.opt_in}"
        )
        event_trace = EventTrace(
            task_id,
            self.paper_finder.retriever.n_retrieval if hasattr(self.paper_finder.retriever, "n_retrieval") else 0,
            # noqa
            self.paper_finder.n_rerank,
            req,
            user_id=user_id
        )
        cost_args = CostReportingArgs(
            task_id=task_id,
            user_id=user_id,
            description="Step 0: Query decomposition",
            model=self.llm_model,
            msg_id=msg_id
        )
        llm_processed_query = self.preprocess_query(query, cost_args)
        event_trace.trace_decomposition_event(llm_processed_query)

        # Paper finder step - retrieve relevant paper passages from semantic scholar index and api
        snippet_srch_res, s2_srch_res = self.find_relevant_papers(llm_processed_query.result)
        retrieved_candidates = snippet_srch_res + s2_srch_res
        if not retrieved_candidates:
            raise Exception(
                f"There is no relevant information in the retrieved snippets for query: {query}.")
        event_trace.trace_retrieval_event(retrieved_candidates)

        # Rerank the retrieved candidates based on the query with a cross encoder
        s2_srch_metadata = [{k: v for k, v in paper.items() if
                             k == "corpus_id" or k in NUMERIC_META_FIELDS or k in CATEGORICAL_META_FIELDS} for paper in
                            s2_srch_res]
        reranked_df, paper_metadata = self.rerank_and_aggregate(query, retrieved_candidates,
                                                                {str(paper["corpus_id"]): paper for paper in
                                                                 s2_srch_metadata})
        if reranked_df.empty:
            raise Exception(
                "No relevant papers found for the query post reranking, skipping quote extraction.")
        event_trace.trace_rerank_event(reranked_df.to_dict(orient="records"))

        # Step 1 - quote extraction
        per_paper_summaries = self.step_select_quotes(query, reranked_df, cost_args)
        if not per_paper_summaries.result:
            raise Exception(
                "No relevant quotes extracted for the query, can't proceed further.")
        event_trace.trace_quote_event(per_paper_summaries)

        # step 2: outline planning and clustering
        cluster_json = self.step_clustering(query, per_paper_summaries.result, cost_args)
        # Changing to expected format in the summary generation prompt
        plan_json = {f'{dim["name"]} ({dim["format"]})': dim["quotes"] for dim in cluster_json.result["dimensions"]}
        if not any([len(d) for d in plan_json.values()]):
            raise Exception("The planning step failed to cluster the relevant documents.")
        event_trace.trace_clustering_event(cluster_json, plan_json)

        # step 2.1: extend the clustered snippets with their inline citations
        per_paper_summaries_extd, quotes_metadata = self.extract_quote_citations(reranked_df,
                                                                                 per_paper_summaries.result,
                                                                                 plan_json, paper_metadata)
        event_trace.trace_inline_citation_following_event(per_paper_summaries_extd, quotes_metadata)

        # step 3: generating output as per the outline
        section_titles = [dim["name"] for dim in cluster_json.result["dimensions"]]
        gen_sections_iter = self.step_gen_iterative_summary(query, per_paper_summaries_extd,
                                                            plan_json, cost_args)

        json_summary, generated_sections, table_threads = [], [], []
        tables = [None for _ in cluster_json.result["dimensions"]]
        citation_ids = dict()

        task_estimated_time = 30 + 15 * len(plan_json)
        task_estimated_time = max((task_estimated_time + task_estimated_time % 60) // 60, 1)
        outline = '\n    - ' + '\n    - '.join(section_titles)
        self.update_task_state(f"Start generating each section in the answer outline: {outline}",
                               task_estimated_time=f"~{task_estimated_time} minutes" if task_estimated_time > 1 else "~1 minute",
                               step_estimated_time=15)

        try:
            gen_iter = gen_sections_iter
            idx = 0
            while True:
                if idx < len(plan_json):
                    self.update_task_state(
                        f"Iteratively generating section: {(idx + 1)} of {len(plan_json)} - {section_titles[idx]}",
                        curr_response=generated_sections, step_estimated_time=15)
                section_text = next(gen_iter)
                section_json = \
                    get_json_summary(self.multi_step_pipeline.llm_model, [section_text], per_paper_summaries_extd,
                                     paper_metadata,
                                     citation_ids, inline_tags)[0]
                section_json["format"] = cluster_json.result["dimensions"][idx]["format"]

                json_summary.append(section_json)
                self.postprocess_json_output(json_summary, quotes_meta=quotes_metadata)
                if section_json["format"] == "list" and section_json["citations"] and self.run_table_generation:
                    cluster_json.result["dimensions"][idx]["idx"] = idx
                    cit_ids = [int(c["paper"]["corpus_id"]) for c in section_json["citations"]]
                    tthread = self.gen_table_thread(user_id, query, cluster_json.result["dimensions"][idx], cit_ids,
                                                    tables)
                    if tthread:
                        table_threads.append(tthread)
                gen_sec = self.get_gen_sections_from_json(section_json)
                generated_sections.append(gen_sec)
                idx += 1
        except StopIteration as e:
            all_sections = e.value

        self.update_task_state(f"Generating comparison tables", curr_response=generated_sections,
                               step_estimated_time=20)

        start = time()
        for tthread in table_threads:
            tthread.join()
        logger.info(f"Adhoc Table generation wait time: {time() - start:.2f}")
        tcosts = []
        for sidx in range(len(json_summary)):
            tables_val = None
            if tables[sidx]:
                if type(tables[sidx]) == tuple:
                    tables_val, tcost = tables[sidx]
                    tcosts.append(tcost)
                else:
                    tables_val = tables[sidx]
            json_summary[sidx]["table"] = tables_val.to_dict() if tables_val else None
            generated_sections[sidx].table = tables_val if tables_val else None
        self.postprocess_json_output(json_summary, quotes_meta=quotes_metadata)
        event_trace.trace_summary_event(json_summary, all_sections, tcosts)
        event_trace.persist_trace(self.logs_config)
        return TaskResult(sections=generated_sections, cost=event_trace.total_cost, tokens=event_trace.tokens)



from abc import ABC, abstractmethod
from typing import List, Any, Dict
import logging

from scholarqa.utils import query_s2_api, METADATA_FIELDS, make_int, NUMERIC_META_FIELDS

logger = logging.getLogger(__name__)


class AbstractRetriever(ABC):
    @abstractmethod
    def retrieve_passages(self, query: str, **filter_kwargs) -> List[Dict[str, Any]]:
        pass

    @abstractmethod
    def retrieve_additional_papers(self, query: str, **filter_kwargs) -> List[Dict[str, Any]]:
        pass


class FullTextRetriever(AbstractRetriever):
    def __init__(self, n_retrieval: int = 256, n_keyword_srch: int = 20):
        self.n_retrieval = n_retrieval
        self.n_keyword_srch = n_keyword_srch

    def retrieve_passages(self, query: str, **filter_kwargs) -> List[Dict[str, Any]]:
        """Query the Semantic Scholar API full text search end point to retrieve papers based on the query.
        The full text search end point does not return the required metadata fields, so we will fetch metadata for
        these later."""
        snippets_list = self.snippet_search(query, **filter_kwargs)
        snippets_list = [
            snippet for snippet in snippets_list if len(snippet["text"].split(" ")) > 20
        ]
        return snippets_list

    def snippet_search(self, query: str, **filter_kwargs) -> List[Dict[str, Any]]:
        if not self.n_retrieval:
            return []
        query_params = {fkey: fval for fkey, fval in filter_kwargs.items() if fval}
        query_params.update({"query": query, "limit": self.n_retrieval})
        print(query_params)
        snippets = query_s2_api(
            end_pt="snippet/search",
            params=query_params,
            method="get",
        )
        snippets_list = []
        res_data = snippets["data"]
        if res_data:
            for fields in res_data:
                res_map = dict()
                snippet, paper = fields["snippet"], fields["paper"]
                res_map["corpus_id"] = paper["corpusId"]
                res_map["title"] = paper["title"]
                res_map["text"] = snippet["text"]
                res_map["score"] = fields["score"]
                res_map["section_title"] = snippet["snippetKind"]
                if snippet["snippetKind"] == "body":
                    if section := snippet.get("section"):
                        res_map["section_title"] = section
                if "snippetOffset" in snippet and snippet["snippetOffset"].get("start"):
                    res_map["char_start_offset"] = snippet["snippetOffset"]["start"]
                else:
                    res_map["char_start_offset"] = 0
                if "annotations" in snippet and "sentences" in snippet["annotations"] and snippet["annotations"][
                    "sentences"]:
                    res_map["sentence_offsets"] = snippet["annotations"]["sentences"]
                else:
                    res_map["sentence_offsets"] = []

                if snippet.get("annotations") and snippet["annotations"].get("refMentions"):
                    res_map["ref_mentions"] = [rmen for rmen in
                                               snippet["annotations"]["refMentions"] if rmen.get("matchedPaperCorpusId")
                                               and rmen.get("start") and rmen.get("end")]
                else:
                    res_map["ref_mentions"] = []
                res_map["pdf_hash"] = snippet.get("extractionPdfHash","")
                res_map["stype"] = "vespa"
                if res_map:
                    snippets_list.append(res_map)
        return snippets_list

    def retrieve_additional_papers(self, query: str, **filter_kwargs) -> List[Dict[str, Any]]:
        return self.keyword_search(query, **filter_kwargs) if self.n_keyword_srch else []

    def keyword_search(self, kquery: str, **filter_kwargs) -> List[Dict[str, Any]]:
        """Query the Semantic Scholar API keyword search end point and return top n papers.
        The keyword search api also accepts filters for fields like year, venue, etc. which we obtain after decomposing
        the initial user query. This end point returns the required metadata fields as well, so we will skip fetching
        metadata for these later."""

        paper_data = []
        query_params = {fkey: fval for fkey, fval in filter_kwargs.items() if fval}
        query_params.update({"query": kquery, "limit": self.n_keyword_srch, "fields": METADATA_FIELDS})
        res = query_s2_api(
            end_pt="paper/search",
            params=query_params,
            method="get",
        )
        if "data" in res:
            paper_data = res["data"]
            paper_data = [pd for pd in paper_data if pd.get("corpusId") and pd.get("title") and pd.get("abstract")]
            paper_data = [{k: make_int(v) if k in NUMERIC_META_FIELDS else pd.get(k) for k, v in pd.items()}
                          for pd in paper_data]
            for pd in paper_data:
                pd["corpus_id"] = str(pd["corpusId"])
                pd["text"] = pd["abstract"]
                pd["section_title"] = "abstract"
                pd["char_start_offset"] = 0
                pd["sentence_offsets"] = []
                pd["ref_mentions"] = []
                pd["score"] = 0.0
                pd["stype"] = "public_api"
                pd["pdf_hash"] = ""
        return paper_data



from sentence_transformers import CrossEncoder


class PaddedCrossEncoder(CrossEncoder):
    def smart_batching_collate_text_only(self, batch):
        texts = [[] for _ in range(len(batch[0]))]

        for example in batch:
            for idx, text in enumerate(example):
                texts[idx].append(text.strip())

        tokenized = self.tokenizer(*texts, padding="max_length", truncation='longest_first', return_tensors="pt",
                                   max_length=self.max_length)

        for name in tokenized:
            tokenized[name] = tokenized[name].to(self._target_device)

        return tokenized


import json
import logging
from typing import Callable, Literal

from pydantic import BaseModel, Field

from scholarqa.state_mgmt.local_state_mgr import AbsStateMgrClient
from scholarqa.utils import TaskIdAwareLogFormatter, init_settings

logger = logging.getLogger(__name__)


class LogsConfig(BaseModel):
    log_dir: str = Field(default="logs", description="Directory to store logs, event traces and litellm cache")
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = Field(default="INFO",
                                                                                description="Logging level")
    llm_cache_dir: str = Field(default="llm_cache", description="Sub directory to cache llm calls")
    event_trace_loc: str = Field(default="scholarqa_traces", description="Sub directory to store event traces"
                                                                         "OR the GCS bucket name")
    extras: dict = Field(default=None, description="Extra arguments for the logs")
    tracing_mode: Literal["local", "gcs"] = Field(default="local",
                                                  description="Mode to store event traces (local or gcs)")
    tid_log_formatter: TaskIdAwareLogFormatter = Field(default=None,
                                                       description="Task Id aware log formatter which prepends the current task id to every log message")

    @property
    def task_id(self):
        return self.tid_log_formatter.task_id

    @task_id.setter
    def task_id(self, task_id: str):
        self.tid_log_formatter.task_id = task_id

    def init_formatter(self):
        self.tid_log_formatter = init_settings(logs_dir=self.log_dir, log_level=self.log_level,
                                               litellm_cache_dir=self.llm_cache_dir)

    class Config:
        arbitrary_types_allowed = True


class RunConfig(BaseModel):
    retrieval_service: str = Field(default="public_api", description="Service to use for paper retrieval")
    retriever_args: dict = Field(default=None, description="Arguments for the retrieval service")
    reranker_service: str = Field(default="modal", description="Service to use for paper reranking")
    reranker_args: dict = Field(default=None, description="Arguments for the reranker service")
    paper_finder_args: dict = Field(default=None, description="Arguments for the paper finder service")
    pipeline_args: dict = Field(default=None, description="Arguments for the Scholar QA pipeline service")


class AppConfig(BaseModel):
    logs: LogsConfig = Field(default=None, description="Configuration for logs and event traces")
    run_config: RunConfig = Field(default=None, description="Configuration for components of the ScholarQA pipeline")
    state_mgr_client: AbsStateMgrClient = Field(default=None,
                                                description="State manager client for managing async task states and cost reporting of llm calls")
    load_scholarqa: Callable = Field(default=None,
                                     description="Function to lazy load ScholarQA instance, should be overridden by the app using this config class")

    class Config:
        arbitrary_types_allowed = True


def read_json_config(config_path: str, model_class: AppConfig = AppConfig) -> AppConfig:
    logger.info(f"Reading app config from {config_path}")
    json_data = json.load(open(config_path, "r"))
    app_config = model_class.model_validate(json_data)
    app_config.logs.tid_log_formatter = init_settings(logs_dir=app_config.logs.log_dir,
                                                        log_level=app_config.logs.log_level,
                                                      litellm_cache_dir=app_config.logs.llm_cache_dir)
    return app_config
